from production.box import SkipCourseBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class SkipCourseBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(SkipCourseBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> SkipCourseBoxBuilder
        return SkipCourseBox(config.name,
                             config.sons,
                             self.get_extra(config, "SkipPodTypes", []),
                             self.get_extra(config, "SkipSaleTypes", []),
                             self.get_logger(config))



from typing import List
from messageprocessorutil import UuidMessageProcessor
from messagebus import NakMessage, AckMessage

from production.interactor import GetPrintersInfoInteractor
from production.box import ViewBox


class GetPrintersInfoProcessor(UuidMessageProcessor):
    def __init__(self, view_boxes):
        # type: (List[ViewBox]) -> None
        super(GetPrintersInfoProcessor, self).__init__()
        self.interactor = GetPrintersInfoInteractor(view_boxes)

    def parse_data(self, _):
        return ''

    def call_business(self, obj):
        return self.interactor.execute()

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, AckMessage(result))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage())

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage())



# -*- coding: utf-8 -*-

# enum.py
# Part of ‘enum’, a package providing enumerated types for Python.
#
# Copyright © 2007–2018 Ben Finney <ben+python@benfinney.id.au>
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3 of that license or any later version.
# No warranty expressed or implied. See the file ‘LICENSE.GPL-3’ for details.

""" Robust enumerated type support in Python.

This package provides a module for robust enumerations in Python.

An enumeration object is created with a sequence of string arguments
to the Enum() constructor::

    >>> from enum import Enum
    >>> Colours = Enum('red', 'blue', 'green')
    >>> Weekdays = Enum('mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun')

The return value is an immutable sequence object with a value for each
of the string arguments. Each value is also available as an attribute
named from the corresponding string argument::

    >>> pizza_night = Weekdays[4]
    >>> shirt_colour = Colours.green

The values are constants that can be compared only with values from
the same enumeration; comparison with other values will invoke
Python's fallback comparisons::

    >>> pizza_night == Weekdays.fri
    True
    >>> shirt_colour > Colours.red
    True
    >>> shirt_colour == "green"
    False

Each value from an enumeration exports its sequence index
as an integer, and can be coerced to a simple string matching the
original arguments used to create the enumeration::

    >>> str(pizza_night)
    'fri'
    >>> shirt_colour.index
    2

"""


__author_name__ = "Ben Finney"
__author_email__ = "ben+python@benfinney.id.au"
__author__ = "%(__author_name__)s <%(__author_email__)s>" % vars()

_copyright_year_begin = "2007"
__date__ = "2018-08-22"
_copyright_year_latest = __date__.split('-')[0]
_copyright_year_range = _copyright_year_begin
if _copyright_year_latest > _copyright_year_begin:
    _copyright_year_range += "–%(_copyright_year_latest)s" % vars()
__copyright__ = (
    "Copyright © %(_copyright_year_range)s"
    " %(__author_name__)s") % vars()
__license__ = "GPL-3.0+"

__url__ = "https://pypi.org/project/enum/"
__version__ = "0.4.7"


class EnumException(Exception):
    """ Base class for all exceptions in this module. """

    def __init__(self, *args, **kwargs):
        if self.__class__ is EnumException:
            class_name = self.__class__.__name__
            raise NotImplementedError(
                "%(class_name)s is an abstract base class" % vars())
        super(EnumException, self).__init__(*args, **kwargs)


class EnumEmptyError(AssertionError, EnumException):
    """ Raised when attempting to create an empty enumeration. """

    def __str__(self):
        return "Enumerations cannot be empty"


class EnumBadKeyError(TypeError, EnumException):
    """ Raised when creating an Enum with non-string keys. """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return "Enumeration keys must be strings: %(key)r" % vars(self)


class EnumImmutableError(TypeError, EnumException):
    """ Raised when attempting to modify an Enum. """

    def __init__(self, *args):
        self.args = args

    def __str__(self):
        return "Enumeration does not allow modification"


def _comparator(func):
    """ Decorator for EnumValue rich comparison methods. """
    def comparator_wrapper(self, other):
        try:
            assert self.enumtype == other.enumtype
            result = func(self.index, other.index)
        except (AssertionError, AttributeError):
            result = NotImplemented

        return result
    comparator_wrapper.__name__ = func.__name__
    comparator_wrapper.__doc__ = getattr(float, func.__name__).__doc__
    return comparator_wrapper

class EnumValue(object):
    """ A specific value of an enumerated type. """

    def __init__(self, enumtype, index, key):
        """ Set up a new instance. """
        self._enumtype = enumtype
        self._index = index
        self._key = key

    @property
    def enumtype(self):
        return self._enumtype

    @property
    def key(self):
        return self._key

    def __str__(self):
        return str(self.key)

    @property
    def index(self):
        return self._index

    def __repr__(self):
        return "EnumValue(%(_enumtype)r, %(_index)r, %(_key)r)" % vars(self)

    def __hash__(self):
        return hash(self._index)

    @_comparator
    def __eq__(self, other):
        return (self == other)

    @_comparator
    def __ne__(self, other):
        return (self != other)

    @_comparator
    def __lt__(self, other):
        return (self < other)

    @_comparator
    def __le__(self, other):
        return (self <= other)

    @_comparator
    def __gt__(self, other):
        return (self > other)

    @_comparator
    def __ge__(self, other):
        return (self >= other)


class Enum(object):
    """ Enumerated type. """

    def __init__(self, *keys, **kwargs):
        """ Create an enumeration instance. """

        value_type = kwargs.get('value_type', EnumValue)

        if not keys:
            raise EnumEmptyError()

        keys = tuple(keys)
        values = [None] * len(keys)

        for i, key in enumerate(keys):
            value = value_type(self, i, key)
            values[i] = value
            try:
                super(Enum, self).__setattr__(key, value)
            except TypeError:
                raise EnumBadKeyError(key)

        self.__dict__['_keys'] = keys
        self.__dict__['_values'] = values

    def __setattr__(self, name, value):
        raise EnumImmutableError(name)

    def __delattr__(self, name):
        raise EnumImmutableError(name)

    def __len__(self):
        return len(self._values)

    def __getitem__(self, index):
        return self._values[index]

    def __setitem__(self, index, value):
        raise EnumImmutableError(index)

    def __delitem__(self, index):
        raise EnumImmutableError(index)

    def __iter__(self):
        return iter(self._values)

    def __contains__(self, value):
        is_member = False
        if isinstance(value, basestring):
            is_member = (value in self._keys)
        else:
            is_member = (value in self._values)
        return is_member


# Local variables:
# mode: python
# time-stamp-format: "%:y-%02m-%02d"
# time-stamp-start: "__date__ = \""
# time-stamp-end: "\"$"
# time-stamp-line-limit: 200
# End:
# vim: filetype=python fileencoding=utf-8 :



from customer_services.interactor.get_customer.dto import GetCustomerRequest
from customer_services.model import Customer
from customer_services.service import GetCustomerService


class GetCustomerInteractor(object):
    def __init__(self, get_customer_service):
        # type: (GetCustomerService) -> None
        
        self.get_customer_service = get_customer_service

    def get_customer(self, request):
        # type: (GetCustomerRequest) -> Customer

        return self.get_customer_service.get_customer(request.phone)



from logging import Logger

from model.config import Config
from model.electronic_message import ElectronicMessage
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.repository import PosRepositoryModel
from model.services import FiscalServiceModel
from interactor.route.electronic_payment.read_messages.dto import (
    ReadMessagesRequest,
    ReadMessagesResponse
)
from validation_util import validate_api_key, check_if_business_day_is_opened, \
    check_if_operator_is_logged, check_if_order_is_totaled
from typing import List


class ReadMessagesInteractor(object):
    def __init__(self, logger, config, pos_repository, fiscal_service, electronic_messages):
        # type: (Logger, Config, PosRepositoryModel, FiscalServiceModel, List[ElectronicMessage]) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.fiscal_service = fiscal_service
        self.electronic_messages = electronic_messages

    def execute(self, request):
        # type: (ReadMessagesRequest) -> ReadMessagesResponse

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        if not pos_id:
            raise ValidationError(PosValidationErrorsType.PosIdNotFound)

        model = self.pos_repository.get_model(pos_id)

        check_if_business_day_is_opened(self.pos_repository, model)
        check_if_operator_is_logged(self.pos_repository, model)
        check_if_order_is_totaled(self.pos_repository, model)

        pos_id = int(pos_id)
        self._remove_read_messages(pos_id, request)

        response_electronic_messages = self._get_response_messages(pos_id, request)

        return response_electronic_messages

    def _remove_read_messages(self, pos_id, request):
        # type: (int, ReadMessagesRequest) -> None

        for message_id in request.read_messages_id:
            for electronic_message in self.electronic_messages[:]:
                if electronic_message.message_id == message_id and electronic_message.pos_id == pos_id:
                    self.electronic_messages.remove(electronic_message)

    def _get_response_messages(self, pos_id, request):
        # type: (int, ReadMessagesRequest) -> List[ElectronicMessage]

        response_messages = []
        for electronic_message in self.electronic_messages:
            if int(electronic_message.pos_id) == int(pos_id):
                electronic_message.pos_id = int(request.pos_id)
                response_messages.append(electronic_message)
        return response_messages



from typing import List


class OrderDetail(object):

    def __init__(self, order_id=None, sale_type=None, state=None, total_amount=None, total_tender=None, due_amount=None,
                 created_at=None, state_history=None, tenders=None, additional_properties=None, items=None):
        # type: (int, str, str, float, float, float, str, List[StateHistory], List[Tender], List[AdditionalProperty], List[Item]) -> None

        self.order_id = order_id
        self.sale_type = sale_type
        self.state = state
        self.total_amount = total_amount
        self.total_tender = total_tender
        self.due_amount = due_amount
        self.created_at = created_at
        self.state_history = state_history
        self.tenders = tenders
        self.additional_properties = additional_properties
        self.items = items


class StateHistory(object):

    def __init__(self, state_name=None, timestamp=None):
        # type: (str, str) -> None

        self.state_name = state_name
        self.timestamp = timestamp


class Tender(object):

    def __init__(self, sequence_id=None, type_id=None, type_name=None, timestamp=None, amount=None):
        # type: (int, int, str, str, float) -> None

        self.sequence_id = sequence_id
        self.type_id = type_id
        self.type_name = type_name
        self.timestamp = timestamp
        self.amount = amount


class AdditionalProperty(object):

    def __init__(self, key=None, value=None):
        # type: (str, str) -> None

        self.key = key
        self.value = value


class Item(object):

    def __init__(self, context=None, level=None, line_number=None, part_code=None, product_name=None, item_type=None,
                 added_price=None, unit_price=None, quantity=None, default_quantity=None):
        # type: (str, int, int, int, str, str, float, float, int, int) -> None

        self.context = context
        self.level = level
        self.line_number = line_number
        self.part_code = part_code
        self.product_name = product_name
        self.item_type = item_type
        self.added_price = added_price
        self.unit_price = unit_price
        self.quantity = quantity
        self.default_quantity = default_quantity



from mw_helper import ensure_iterable

from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import Item, ProductionOrder
from typing import Union


class ProductCodeFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, product_codes, keep_children=False, logger=None):
        super(ProductCodeFilterBox, self).__init__(name, sons, logger)
        self.product_codes = ensure_iterable(product_codes)
        self.keep_children = keep_children

    def change_order(self, order):
        self._handle_items(order)

        return order

    def _handle_items(self, item):
        # type: (Union[ProductionOrder, Item]) -> None
        def verify(product):
            return unicode(product.part_code) not in self.product_codes

        new_order_items = []
        for current_item in item.items:
            if verify(current_item):
                new_order_items.append(current_item)

                if current_item.items:
                    self._handle_items(current_item)

            elif str(self.keep_children).lower() == "true":
                for child in current_item.items:
                    child.level = int(child.level) - 1
                    child.qty *= current_item.qty

                    if not verify(child):
                        continue

                    new_order_items.append(child)
                    self._handle_items(child)
        item.items = new_order_items



from logging import Logger
from typing import List

from interactor.route.menu.get_navigation.dto import GetNavigationRequest
from model.config import Config
from model.error_type import ExternalServicesErrorsType
from model.exception import ExternalServiceErrors
from model.repository import PosRepositoryModel
from validation_util import validate_api_key


class GetNavigationInteractor(object):
    def __init__(
            self,
            logger,             # type: Logger
            config,             # type: Config
            pos_repository,     # type: PosRepositoryModel
            cache_navigation    # type: List
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.cache_menu_navigation = cache_navigation

    def execute(
        self,
        request     # type: GetNavigationRequest
    ):
        # type: (...) -> dict

        pos_id = validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping
        )

        navigation = None
        if not self.cache_menu_navigation:
            navigation = self.pos_repository.get_navigation(pos_id)

        if navigation:
            self.cache_menu_navigation.append(navigation)

        if not self.cache_menu_navigation:
            raise ExternalServiceErrors(ExternalServicesErrorsType.MenuNavigationNotAvailable)

        return self.cache_menu_navigation[0]



from _CashTransferReportDto import CashTransferReportDto
from _TransfersByDayDto import TransfersByDayDto
from _Transfer import Transfer
from _CashReportFiltersDto import CashReportFiltersDto

__all__ = ["CashTransferReportDto",
           "TransfersByDayDto",
           "Transfer",
           "CashReportFiltersDto"]



class LogisticPartnerNotFound(Exception):
    pass


class LogisticSearchError(Exception):
    pass


class LogisticCancelError(Exception):
    pass


class LogisticRequestError(Exception):
    pass


class SaveDeliveryOrderError(Exception):
    pass


class SendLogisticError(Exception):
    pass


class SendOrderToProductionError(Exception):
    pass


class DeliveryCouponPrintError(Exception):
    pass


class NotFoundUserError(Exception):
    pass


class GettingPosVersionException(Exception):
    pass


class BenefitAppliedException(Exception):
    pass


class RemoveBenefitException(Exception):
    pass


class ErrorGettingBenefits(Exception):
    pass


class ErrorAddingBenefits(Exception):
    pass


class InvalidTipError(Exception):
    pass


class UpdateRuptureException(Exception):
    pass


class ErrorGettingSalesChallenge(Exception):
    pass


class ErrorOnInsertSalesChallenge(Exception):
    pass


class ErrorOnUpdateSalesChallenge(Exception):
    pass


class ErrorOnDeleteSalesChallenge(Exception):
    pass


class DateFormatException(Exception):
    pass


class GettingUserInfoException(Exception):
    pass


class CanceledDialogException(Exception):
    pass


class RequiredDocumentException(Exception):
    pass


class ErrorGettingNFType(Exception):
    pass


class ErrorGettingCertificateExpirationDays(Exception):
    pass


class ErrorApplyingDiscount(Exception):
    pass


class DiscountMinLevelException(Exception):
    pass


class GetRuptureItemsException(Exception):
    pass


class MinQuantityException(Exception):
    pass


class MaxQuantityException(Exception):
    pass


class ElectronicPaymentException(Exception):
    pass


class ProductNotExistsException(Exception):
    pass




from typing import Optional

from httpmessageprocessorutil.automapped import auto_map


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "payment_id": {"name": "paymentId", "type": int},
           "amount": {"name": "amount", "type": float},
           "is_register_payment": {"name": "isRegisterPayment", "type": Optional[bool]},
           "id_auth": {"name": "idAuth", "type": Optional[str]},
           "media": {"name": "media", "type": Optional[str]},
           "original_nfce_brand": {"name": "originalNfceBrand", "type": Optional[str]},
           "nfce_brand": {"name": "nfceBrand", "type": Optional[str]},
           "receipt_merchant": {"name": "receiptMerchant", "type": Optional[str]},
           "receipt_customer": {"name": "receiptCustomer", "type": Optional[str]},
           "transaction_processor": {"name": "transactionProcessor", "type": Optional[str]},
           "cnpj_auth": {"name": "cnpjAuth", "type": Optional[str]},
           "brand": {"name": "bandeira", "type": Optional[str]},
           "auth_Code": {"name": "authCode", "type": Optional[str]},
           "transaction_identifier": {"name": "transactionIdentifier", "type": Optional[str]},
           "card_number": {"name": "cardNumber", "type": Optional[str]},
           "accreditor_sat": {"name": "accreditorSat", "type": Optional[str]},
           "nsu_host": {"name": "nsuHost", "type": Optional[str]}})
class OrderInsertPaymentRequest(object):
    def __init__(
        self,
        system_api_key,                 # type: str
        pos_id,                         # type: str
        payment_id,                     # type: int
        amount,                         # type: float
        is_register_payment=False,      # type: Optional[bool]
        id_auth=None,                   # type: Optional[str]
        media=None,                     # type: Optional[str]
        original_nfce_brand=None,       # type: Optional[str]
        nfce_brand=None,                # type: Optional[str]
        receipt_merchant=None,          # type: Optional[str]
        receipt_customer=None,          # type: Optional[str]
        transaction_processor=None,     # type: Optional[str]
        cnpj_auth=None,                 # type: Optional[str]
        brand=None,                     # type: Optional[str]
        auth_code=None,                 # type: Optional[str]
        transaction_identifier=None,    # type: Optional[str]
        card_number=None,               # type: Optional[str]
        accreditor_sat=None,            # type: Optional[str]
        nsu_host=None,                  # type: Optional[str]
    ):
        # type: (...) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.payment_id = payment_id
        self.electronic_type_id = payment_id
        self.amount = amount
        self.is_register_payment = is_register_payment
        self.id_auth = id_auth
        self.media = media
        self.original_nfce_brand =original_nfce_brand
        self.nfce_brand = nfce_brand
        self.receipt_merchant = receipt_merchant
        self.receipt_customer = receipt_customer
        self.transaction_processor = transaction_processor
        self.cnpj_auth = cnpj_auth
        self.brand = brand
        self.auth_code = auth_code
        self.transaction_identifier = transaction_identifier
        self.card_number = card_number
        self.accreditor_sat = accreditor_sat
        self.nsu_host = nsu_host



class OperatorGoals(object):
    def __init__(self, operator_id, amount_goal, amount_sold, average_ticket, item_goals, tickets):
        self.id = operator_id
        self.amount_goal = amount_goal
        self.amount_sold = amount_sold
        self.average_ticket = average_ticket
        self.item_goals = item_goals
        self.tickets = tickets



import operator
import iso8601

from production.model import ProductionOrder, StateEvent
from typing import List

from ._OrderChangerProductionBox import OrderChangerProductionBox


def get_time_diff_from_timestamp(first_time, last_time):
    return (iso8601.parse_date(last_time) - iso8601.parse_date(first_time)).total_seconds()


def _get_consolidated_state_history(order):
    # type: (ProductionOrder) -> List[StateEvent]
    return sorted(list(order.state_history) + list(order.prod_state_history), key=operator.attrgetter("timestamp"))


class DecreaseTimeTagBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger, get_config):
        super(DecreaseTimeTagBox, self).__init__(name, sons, logger)
        self.first_state_value = get_config("FirstState").get("State")
        self.first_state_type = get_config("FirstState").get("Type")
        self.last_state_value = get_config("LastState").get("State")
        self.last_state_type = get_config("LastState").get("Type")
        self.tag_name = get_config("TagName")

    def change_order(self, order):
        if self.tag_name not in order.tags:
            return order
        time_to_decrease = 0

        found_start_state = None
        found_end_state = None
        for state in _get_consolidated_state_history(order):
            if not found_start_state:
                found_start_state = self._check_found_state(state, self.first_state_type, self.first_state_value)

            elif not found_end_state:
                found_end_state = self._check_found_state(state, self.last_state_type, self.last_state_value)

            else:
                time_to_decrease += get_time_diff_from_timestamp(found_start_state.timestamp, found_end_state.timestamp)
                found_start_state = found_end_state = None

        order.tags[self.tag_name] -= time_to_decrease

        return order

    @staticmethod
    def _check_found_state(state, state_type, state_value):
        found_state = None
        if state_type.lower() == "order_state":
            if state.state == state_value:
                found_state = state

        elif state.prod_state == state_value:
            found_state = state

        return found_state



import json
import logging
import time

from datetime import datetime, timedelta
from logging import Logger
from threading import Lock, Condition, Thread, local
from xml.etree import cElementTree as eTree

from mw_helper import ensure_list
from production.box import ProductionBox
from production.command import ProductionCommandProcessor, ToggleTagLinesCommand, ChangeProdStateCommand, \
    UndoProductionCommand, AddTagLinesCommand, ChangePathCommand, OrderModifiedCommand, PublishOrderCommand, \
    ProductionCommand, PurgeOrdersCommand
from production.command.undo import UndoCommand, TagLinesUndoCommand, ChangeProdStateUndoCommand, \
    TagLinesUndoCommandType
from production.model import ProductionOrder, ProdStates
from production.repository import ProductionRepository
from production.view import OrderXml, KdsView
from messagebus import MessageBus, Message, DataType, TokenPriority, TokenCreator, Event
from typing import List, Dict, Any, Optional, Callable
from util import handle_datetime_str

from ._CommandQueue import CommandQueue
from ._OrderChanger import OrderChanger
from ._PublishScheduler import PublishScheduler
from ._StatisticUpdater import StatisticUpdater

performance_event_logger = logging.getLogger("ProductionPerformanceCommand")


MSGGRP_KDS = '0B'
TK_KDS_UPDATE_STATISTICS = TokenCreator.create_token(TokenPriority.high, MSGGRP_KDS, "21")


class ProductionManager(ProductionCommandProcessor, PublishScheduler, OrderChanger, StatisticUpdater):
    def __init__(
            self,
            production_repository,  # type: ProductionRepository
            purge_interval,         # type: int
            orders_life_time,       # type: int
            message_bus,            # type: MessageBus
            audit_prod_states,      # type: List
            kds_manager,            # type: str
            logger                  # type: Logger
    ):
        # type: (...) -> None

        self.production_repository = production_repository
        self.purge_interval = purge_interval
        self.orders_life_time = orders_life_time
        self.message_bus = message_bus
        self.logger = logger
        self.audit_logger_prod_states = audit_prod_states
        self.kds_manager = kds_manager

        self.boxes = []  # type: List[ProductionBox]
        self.orders = {}  # type: Dict[str, ProductionOrder]
        self.undo_lock = Lock()
        self.undo_view_map = {}  # type: Dict[str, List[UndoCommand]]
        self.waiters = {}  # type: Dict[str, ProductionManager.PublishWaiter]
        self.waiter_lock = Lock()
        self.purge_waiter = None  # type: Optional[ProductionManager.PurgeWaiter]
        self.statistics = {}  # type: Dict[str, Any]
        self.statistics_changed = local()
        self.statistics_changed.value = None
        self.view_boxes = {}

        self.command_queue = CommandQueue(self, self.logger)

    def remove_tags(self, items):
        for item in items:
            item.tags = {}
            item.tag_history = []

            self.remove_tags(item.items)

    def initialize(self):
        self.purge_waiter = ProductionManager.PurgeWaiter(self, self.logger)
        self.purge_waiter.purge_orders()
        self._wait_purging_orders()

        for order in self.production_repository.get_all_orders():
            order.first_processing = False
            for undo_command in order.undo_list:  # type: UndoCommand
                self.enqueue_undo_command(undo_command.view, undo_command)
            self.orders[order.order_id] = order
            self.sync_publish_order(order)

        self.purge_waiter.purge_orders_and_wait()

    def purge_all_orders(self):
        self.purge_waiter = ProductionManager.PurgeWaiter(self, self.logger)
        self.purge_waiter.purge_orders(all_orders=True)

    @staticmethod
    def _wait_purging_orders():
        time.sleep(5)

    def set_root_boxes(self, boxes):
        # type: (List[ProductionBox]) -> None
        self.boxes = boxes
        for box in self.boxes:
            box.parent = self

    def set_views(self, view_boxes):
        # type: (List) -> None

        for box in view_boxes:
            self.view_boxes[box.get_view_name()] = box

    def process_command(self, command):
        command_name = type(command).__name__
        order_id = "-"
        try:
            start_time = datetime.now()
            tree_box_count = self.inner_process_command(command)

            if command.order_id is not None:
                order_id = command.order_id.split("-")[0]

            self._performance_log(command, command_name, order_id, start_time, tree_box_count)
        except Exception as ex:
            performance_event_logger.exception("Error executing command {} for order {}".format(command_name, order_id))
            raise ex

    @staticmethod
    def _performance_log(command, command_name, order_id, start_time, tree_box_count):
        end_time = datetime.now()
        time_delta = (end_time - start_time).total_seconds()
        message = "{:>25} / {:>10} / {:>8.3f}s".format(command_name, order_id, time_delta)
        if hasattr(command, 'order') and command.order is not None:
            last_order_update_state_str = handle_datetime_str(command.order.state_history[-1].timestamp)
            last_order_update_state = datetime.strptime(last_order_update_state_str, "%Y-%m-%dT%H:%M:%S.%f")
            last_order_update_state_difference = (start_time - last_order_update_state).total_seconds()

            if isinstance(command, OrderModifiedCommand):
                message += " - Info: {:7} / {:>12} / {:>12} / {} / {} / {:>8.3f}s".format(
                    command.order.originator,
                    command.order.state,
                    command.order.prod_state,
                    command.order.recalled,
                    command.order.state_history[-1].timestamp,
                    last_order_update_state_difference,
                )

        if tree_box_count is not None:
            tree_messages = []
            total_boxes = 0
            for key, value in tree_box_count.iteritems():
                tree_messages.append("{} - {}".format(key, str(value)))
                total_boxes += value

            message += " " + " / ".join(tree_messages) + " / " + str(total_boxes)

        performance_event_logger.info(message)

    def inner_process_command(self, command):
        # type: (ProductionCommand) -> Optional[Dict]

        order = None
        if command.order_id is not None:
            command.order_id = command.order_id.split("-")[0]

        if isinstance(command, OrderModifiedCommand):
            tree_box_count = self._handle_order_modified_command(command)
            return tree_box_count

        elif isinstance(command, PublishOrderCommand):
            self.sync_publish_order(command.order)
            return

        elif isinstance(command, PurgeOrdersCommand):
            self.handle_purge_orders_command(command=command)
            return

        elif isinstance(command, ChangePathCommand):
            self._handle_change_path_command(command)
            return

        if command.order_id is not None and command.order_id not in self.orders:
            if command.view in self.view_boxes:
                current_view = self.view_boxes[command.view]
                current_view.delete_order(command.order_id)
                current_view.reload_view()

            return

        if isinstance(command, ToggleTagLinesCommand) or isinstance(command, AddTagLinesCommand):
            order = self.tag_lines(command)

        elif isinstance(command, ChangeProdStateCommand):
            order = self._handle_change_prod_state_command(command)
            self.publish_audit_logger_event(order)

        elif isinstance(command, UndoProductionCommand):
            order = self._handle_undo_production_command(command)
        else:
            self.logger.error("[process_command] Unknown command received: {}".format(type(command)))

        if order is not None:
            self.production_repository.save_order(order)
            self.sync_publish_order(order)

    def _handle_change_path_command(self, command):
        handled = self.publish_command(command)
        for box in handled:
            box.start_refresh()

        for box in handled:
            for order in self.orders.values():
                box.order_modified(order.clone())

        for box in handled:
            box.end_refresh()

    def _handle_undo_production_command(self, command):
        try:
            view_undo_commands = self.undo_view_map[command.view]
            if len(view_undo_commands) == 0:
                return
        except KeyError:
            self.logger.info("No Undo Commands for the View: {}".format(command.view))
            return

        with self.undo_lock:
            next_available = None
            for undo_command in view_undo_commands:
                if undo_command.is_enabled():
                    next_available = undo_command
                    break

            if next_available is not None:
                view_undo_commands.remove(next_available)

        if next_available is None:
            return

        if next_available.order_id not in self.orders:
            self.logger.info("The order {} has no longer available".format(next_available.order_id))
            return

        order = self.orders[next_available.order_id]
        if order is None:
            return
        next_available.undo(order)

        return order

    def _handle_change_prod_state_command(self, command):
        order = self.orders[command.order_id]
        state_undo_command = ChangeProdStateUndoCommand(order.order_id, command.view, order.prod_state, [])
        self.add_undo_command_to_order_and_view(command.view, order, state_undo_command)
        order.prod_state = command.state
        self.statistics_changed.value = False
        self.send_statistics_update()
        return order

    def handle_purge_orders_command(
        self,
        command,    # type: PurgeOrdersCommand
    ):
        # type: (...) -> None

        self.logger.debug("Purging orders command start")
        self.production_repository.purge_orders(command.orders)
        cached_purged_orders = self._get_cached_purged_orders(command.orders)
        self.logger.debug("Syncing cached purged orders")
        for order in cached_purged_orders:
            self.sync_publish_order(order)
            self.orders.pop(order.order_id, None)

    def _handle_order_modified_command(self, command):
        order = command.order
        if order.order_id in self.orders:
            order.first_processing = False
            saved_order = self.orders[order.order_id]
            order.prod_state = saved_order.prod_state
            self.update_order_items(order, saved_order)
            if hasattr(saved_order, "round_robin_path"):
                order.round_robin_path = saved_order.round_robin_path
        else:
            order.first_processing = True

        self.statistics_changed.value = False
        self.orders[order.order_id] = order
        self.production_repository.save_order(order)
        tree_box_count = self.sync_publish_order(order)
        self.send_statistics_update()

        return tree_box_count

    def order_modified(self, order):
        # type: (ProductionOrder) -> None

        self.command_queue.put(OrderModifiedCommand(order.order_id, order))

    def update_order_items(self, order, saved_order):
        # type: (ProductionOrder, ProductionOrder) -> None
        for item in order.items:
            for saved_item in saved_order.items:
                if item.get_line_id() == saved_item.get_line_id():
                    self.add_saved_tags_to_item(item, saved_item)

    def add_saved_tags_to_item(self, item, saved_item):
        item.tags = saved_item.tags
        item.tag_history = saved_item.tag_history
        for son in item.items:
            for saved_son in saved_item.items:
                if son.get_line_id() == saved_son.get_line_id():
                    self.add_saved_tags_to_item(son, saved_son)

    def handle_command(self, command):
        self.command_queue.put(command)

    def schedule_publish(self, order_id, wait_time):
        if order_id is not None:
            order_id = str(order_id)
            order_id = order_id.split("-")[0]

        order = self.orders[order_id]
        publish_waiter = ProductionManager.PublishWaiter(self, order, wait_time)
        start_thread = False
        waiter_to_deactivate = None
        with self.waiter_lock:
            if order.order_id in self.waiters:
                if self.waiters[order.order_id].wait_time > wait_time:
                    waiter_to_deactivate = self.waiters[order.order_id]
                    del self.waiters[order.order_id]
                    self.waiters[order.order_id] = publish_waiter
                    start_thread = True
            else:
                self.waiters[order.order_id] = publish_waiter
                start_thread = True

        if waiter_to_deactivate is not None:
            waiter_to_deactivate.deactivate()

        if start_thread:
            publish_waiter.wait_and_publish()

    def remove_schedule_publish(self, order_id):
        # type: (str) -> None

        if order_id is not None:
            order_id = str(order_id)
            order_id = order_id.split("-")[0]

        if order_id in self.waiters:
            self.waiters[order_id].deactivate()
            del self.waiters[order_id]

    def change_order(self, order_id, order_changer, command_order=None):
        # type: (str, Callable, ProductionOrder) -> None

        if order_id is not None:
            order_id = str(order_id)
        order_id = order_id.split("-")[0]

        if order_id in self.orders:
            order = self.orders[order_id]
            changed_order = order_changer(order, command_order) if command_order else order_changer(order)
            if changed_order:
                self.production_repository.save_order(order)
                self.publish_order(order)

    def set_all_items(self, item_view, item_save, item_setter):
        def get_diff_key(x):
            return '{}-{}-{}'.format(x.line_number, x.item_id, x.part_code)

        if item_save.is_product():
            if get_diff_key(item_save) == get_diff_key(item_view):
                return item_setter(item_save)
            return False

        were_change = False
        for son in item_save.items:
            if self.set_all_items(item_view, son, item_setter):
                were_change = True
        return were_change

    def change_items(self, order, item_changer):
        order_id = order.order_id.split("-")[0]
        saved_order = self.orders[order_id]
        was_change = False

        for view_item in order.items:
            for saved_item in saved_order.items:
                if self.set_all_items(view_item, saved_item, item_changer):
                    was_change = True

        if was_change:
            self.production_repository.save_order(saved_order)

    def tag_lines(self, command):
        order_id = command.order_id.split("-")[0]
        order = self.orders[order_id]
        self.change_order_tags(command, order)
        self.create_undo_command(command, order)
        return order

    @staticmethod
    def change_order_tags(command, order):
        tags = ensure_list(command.tag)

        if isinstance(command, ToggleTagLinesCommand):
            command_callable = order.toggle_tag
        else:
            command_callable = order.add_tag

        for tag in tags:
            for line_id in command.line_ids:
                line_jit_list = command.line_ids_jit_lines.get(line_id) if command.line_ids_jit_lines else None
                command_callable(line_id, tag, line_jit_list)

    def create_undo_command(self, command, order):
        undo_command = self.build_undo_command(command, order)
        self.add_undo_command_to_order_and_view(command.view, order, undo_command)

    def build_undo_command(self, command, order):
        disabled_commands = self.get_disabled_commands(command, order)
        if isinstance(command, ToggleTagLinesCommand):
            undo_type = TagLinesUndoCommandType.toggle
        else:
            undo_type = TagLinesUndoCommandType.add

        undo_command = TagLinesUndoCommand(
            order.order_id,
            command.view,
            command.tag,
            command.line_ids,
            undo_type,
            disabled_commands,
            command.line_ids_jit_lines
        )

        return undo_command

    @staticmethod
    def get_disabled_commands(command, order):
        disabled_commands = []
        if command.tag == "served":
            for undo_command in order.undo_list:
                if isinstance(undo_command, TagLinesUndoCommand) and undo_command.tag in ("doing", "done"):
                    undo_command.disable()
                    disabled_commands.append(undo_command)
        return disabled_commands

    def get_order_xml(self, order_id):
        if order_id not in self.orders:
            return None

        return OrderXml().to_xml(self.orders[order_id])

    def get_production_order(self, order_id):
        if order_id not in self.orders:
            return None

        return self.orders[order_id]

    def publish_order(self, order):
        self.command_queue.put(PublishOrderCommand(order.order_id, order))

    def sync_publish_order(self, order):
        self.logger.info("Publishing order: {}".format(order.order_id))
        with self.waiter_lock:
            if order.order_id in self.waiters:
                self.waiters[order.order_id].deactivate()
                del self.waiters[order.order_id]

        tree_box_count = {}
        for box in self.boxes:
            try:
                self.logger.debug("\n#region    - {} @ {}".format(box.name, box.tree_name))
                cloned_order = order.clone()
                box.order_modified(cloned_order)
                tree_box_count[box.tree_name] = cloned_order.box_count if hasattr(cloned_order, 'box_count') else 1
            except Exception:  # noqa
                self.logger.exception("Exception on box: {}".format(box.name))
            finally:
                self.logger.debug("\n#endregion - {} @ {}".format(box.name, box.tree_name))

        self.logger.info("Finished publishing order: {}".format(order.order_id))

        return tree_box_count

    def publish_command(self, command):
        handled = []
        for box in self.boxes:
            try:
                if box.handle_production_command(command):
                    handled.append(box)
            except Exception as _:  # noqa
                self.logger.exception("Exception on box: {}".format(box.name))

        return handled

    @staticmethod
    def add_tag_to_dict(tags, line_id, tag):
        if line_id not in tags:
            tags[line_id] = []
        tags[line_id].append(tag)

    def add_undo_command_to_order_and_view(self, view, order, undo_command):
        order.undo_list.append(undo_command)
        self.enqueue_undo_command(view, undo_command)

    def enqueue_undo_command(self, view, undo_command):
        with self.undo_lock:
            if view not in self.undo_view_map:
                self.undo_view_map[view] = []
            self.undo_view_map[view].insert(0, undo_command)

    def stop(self):
        if self.purge_waiter is not None:
            self.purge_waiter.deactivate()

    def _get_cached_purged_orders(self, orders_to_purge):
        # type: (List[str]) -> List[ProductionOrder]
        orders = []
        for order_id in orders_to_purge:
            str_order_id = str(order_id)
            if str_order_id in self.orders:
                order = self.orders[str_order_id]
                order.purged = True
                order.prod_state = ProdStates.INVALID
                orders.append(order)
        return orders

    def update_statistics(self, order_id, stats):
        self.statistics_changed.value = True
        self.statistics.update(stats)
        self.send_statistics_update()

    def send_statistics_update(self):
        try:
            message = Message(TK_KDS_UPDATE_STATISTICS, DataType.json, json.dumps(self.statistics), 5000000)
            self.message_bus.send_message("NKDSCTRL", message)
        except BaseException:  # noqa
            pass

    def publish_audit_logger_event(self, order):
        if order.prod_state not in self.audit_logger_prod_states:
            return

        order_id = str(order.order_id)
        try:
            production_order = eTree.fromstring(ProductionManager.get_order_xml(self, order.order_id))
            production_order_xml = eTree.tostring(production_order)

            data = """<ProductionOrderEvent posId="{0}" businessPeriod="{1}" orderId="{2}">{3}</ProductionOrderEvent>"""
            msg_data = data.format(str(order.pos_id), str(order.business_day), order_id, production_order_xml)
            event = Event(subject="PRODUCTION_ORDER_EVENT", event_type="PRODUCTION_ORDER_TIMES", data=msg_data)

            self.logger.info("Publishing AuditLog event: {} / {}".format(event.subject, event.event_type))
            self.message_bus.publish_event(event)
        except Exception as _:  # noqa
            self.logger.exception("Error publishing AuditLogger event for orderId#{}".format(order_id))

    class PublishWaiter(object):
        def __init__(self, production_manager, order, wait_time):
            self.production_manager = production_manager
            self.order = order
            self.wait_time = wait_time
            self.condition = Condition()
            self.active = True

        def wait_and_publish(self):
            t = Thread(target=self._wait_and_publish, name="PublishWaiter")
            t.daemon = True
            t.start()

        def deactivate(self):
            self.active = False
            with self.condition:
                self.condition.notify_all()

        def _wait_and_publish(self):
            with self.condition:
                self.condition.wait(self.wait_time)

            if self.active:
                self.production_manager.publish_order(self.order)

    class PurgeWaiter(object):
        def __init__(self, production_manager, logger):
            self.production_manager = production_manager
            self.logger = logger
            self.wait_time = production_manager.purge_interval
            self.orders_life_time = production_manager.orders_life_time
            self.production_repository = production_manager.production_repository
            self.orders = production_manager.orders
            self.condition = Condition()
            self.active = True

        def deactivate(self):
            self.active = False
            with self.condition:
                self.condition.notify_all()

        def purge_orders_and_wait(self):
            t = Thread(target=self._purge_orders_and_wait, name="PurgeWaiter")
            t.daemon = True
            t.start()

        def _purge_orders_and_wait(self):
            while self.active:
                try:
                    self.logger.debug("Starting thread!")
                    with self.condition:
                        sleep_time = self.wait_time * 60
                        self.logger.debug("Sleeping for {} seconds".format(sleep_time))
                        self.condition.wait(sleep_time)

                    self.purge_orders()
                except BaseException as _:  # noqa
                    self.logger.exception("Error purging old orders")

        def purge_orders(self, all_orders=False):
            self.logger.info("Starting purge process...")

            if not all_orders:
                limit_date = datetime.now() - timedelta(minutes=self.orders_life_time)
                orders_to_purge = self.production_repository.get_orders_to_purge(limit_date)
            else:
                all_orders = self.production_repository.get_all_orders()
                orders_to_purge = [x.order_id for x in all_orders]

            self.logger.info("Found {} orders to purge".format(len(orders_to_purge)))

            if len(orders_to_purge) != 0:
                self.production_manager.handle_command(PurgeOrdersCommand(orders_to_purge))



import json
from datetime import datetime


def to_camel_case(snake_str):
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


def convert_underscore_to_hyphen(snake_str):
    return snake_str.replace("_", "-")


def convert_properties_to_camel_case(obj):
    if isinstance(obj, list):
        ret = []
        for o in obj:
            ret.append(convert_properties_to_camel_case(o))
        return ret
    else:
        d = obj.__dict__
        ret = {}
        for name in d:
            ret[to_camel_case(name)] = d[name]
        return ret


def convert_to_json(obj):
    def mapper(o):
        if isinstance(o, datetime):
            return o.isoformat()
        if hasattr(o, "__dict__"):
            return convert_properties_to_camel_case(o)
        else:
            return str(o)

    if hasattr(obj, "__dict__"):
        obj = convert_properties_to_camel_case(obj)

    return json.dumps(obj, default=mapper)



from logging import Logger
from typing import List, Optional

from mw_helper import convert_to_dict, ensure_iterable
from production.command import CommandType, ToggleTagLinesCommand, ChangeProdStateCommand, UndoProductionCommand, \
    AddTagLinesCommand, ProductionCommandProcessor
from production.manager import OrderChanger
from production.model import create_states_from_state, ProductionOrder, is_product, ProdStates
from production.model.request import ToggleTagLineRequest, ChangeProdStateRequest
from production.view import ProductionView, PrinterView, KdsView, PublishView, CommandView


from ._ProductionBox import ProductionBox


class OrderNotFound(Exception):
    def __init__(self, order_id):
        self.message = "Order {} not found".format(order_id)


class ViewBox(ProductionBox):
    def __init__(self, name, command_processor, order_changer, view, tag_from, prod_state_from, logger=None):
        # type: (str, ProductionCommandProcessor, OrderChanger, ProductionView, str, str, Logger) -> None

        super(ViewBox, self).__init__(name, None, logger)
        self.command_processor = command_processor  # type:
        self.order_changer = order_changer
        self.view = view
        self.can_tag_states = convert_to_dict(create_states_from_state(tag_from))
        self.can_change_prod_state_states = convert_to_dict(create_states_from_state(prod_state_from))

        self.reload_view()

    def order_modified(self, order):
        if self._should_ignore_order(order):
            return

        if isinstance(self.view, KdsView):
            self._handle_kds_view(order)

        if isinstance(self.view, PublishView):
            self._handle_publish_view(order)

        if isinstance(self.view, PrinterView):
            self._handle_printer_view(order)

        if isinstance(self.view, CommandView):
            self._handle_command_view(order)

    def _handle_kds_view(self, order):
        self.save_order(order) if not order.purged else self.delete_order(order.order_id)
        self.debug("Sending order {0} to KDSView: {1}, order:\n{2}", order.order_id, self.view.name, order)
        self.view.handle_order(order)

    def _handle_publish_view(self, order):
        self.debug("Sending order {0} to Publish: {1}", order.order_id, self.view.name)
        self.view.handle_order(order)
        self._tag_order(order)

    def _handle_printer_view(self, order):
        self.debug("Sending order {0} to Printer: {1}", order.order_id, self.view.name)
        self.view.handle_order(order)
        self._tag_order(order)

    def _handle_command_view(self, order):
        self.debug("Sending Command to order: {}".format(order.order_id))
        self.view.handle_order(order)
        self._tag_order(order)

    def _tag_order(self, order):
        tags = self.view.get_view_tags()
        line_ids = self.get_order_line_ids(order)
        if len(line_ids) > 0 and len(tags) > 0:
            def tag_lines(order_to_change, command_order):
                some_tag_changed = False
                for tag in tags:
                    for line_id in line_ids:
                        item = command_order.get_item_by_line_id(line_id)
                        if tag not in item.tags:
                            order_to_change.add_tag(line_id, tag, item.displayed_jit)
                            some_tag_changed = True

                return some_tag_changed

            self.order_changer.change_order(order.order_id, tag_lines, order)

    def _should_ignore_order(self, order):
        # type: (ProductionOrder) -> bool

        stored_order = self.get_order(order.order_id)
        order_is_invalid = order.prod_state == ProdStates.INVALID
        stored_order_is_invalid = not stored_order or stored_order.prod_state == ProdStates.INVALID
        if order_is_invalid and stored_order_is_invalid:
            return True

        return order == stored_order

    def handle_view_command(self, command):
        if command.type == CommandType.toggle_tag_line:
            self._handle_add_tag_line(command)

        elif command.type == CommandType.change_prod_state:
            self._handle_change_prod_state(command)

        elif command.type == CommandType.undo:
            self.command_processor.handle_command(UndoProductionCommand(self.get_view_name()))

        elif command.type == CommandType.refresh_view:
            self._handle_refresh_view()

    def _get_order_with_exception(self, command):
        order = self.get_order(command.data.order_id)
        if not order:
            self.debug("Order {} does not exist. Reloading view...".format(command.data.order_id))
            self.reload_view()
            raise OrderNotFound(command.data.order_id)
        return order

    def _handle_add_tag_line(self, command):
        try:
            order = self._get_order_with_exception(command)
        except OrderNotFound:
            return

        if not self.can_tag(order):
            self.debug("Cannot tag line, order state {}".format(order.state))
            return

        request = command.data  # type: ToggleTagLineRequest
        lines = []
        for sale_line in request.lines:
            for item in order.items:
                if int(sale_line.line_number) != int(item.line_number):
                    continue

                sale_line_key = sale_line.item_id + '.' + sale_line.part_code
                if item.item_code == sale_line_key and not item.block_tags:
                    lines.append(item.get_line_id())
                    for joined_item in item.joined_items:
                        lines.append(joined_item.get_line_id())

        view_name = self.get_view_name()
        production_command = ToggleTagLinesCommand(request.order_id, view_name, lines, request.tag_name, order)
        self.command_processor.handle_command(production_command)

    def _handle_change_prod_state(self, command):
        try:
            order = self._get_order_with_exception(command)
        except OrderNotFound:
            return

        if not self.can_change_prod_state(order):
            self.debug("Cannot change prod state, order state {}".format(order.state))
            return

        request = command.data  # type: ChangeProdStateRequest
        if request.state in ("ITEMS_DONE", "ITEMS_SERVED"):
            order = self.get_order(request.order_id)
            forbidden_tags = ["dont-need-cook"] if request.state == "ITEMS_DONE" else None
            line_ids = self.get_order_line_ids(order, forbidden_tags)

            tag_name = self.get_tag_by_state(request.state)
            production_command = AddTagLinesCommand(request.order_id, self.get_view_name(), line_ids, tag_name, order)

            debug_message = "ViewBox {} tagging with {} lines: {} from order:\n{}"
            self.debug(debug_message.format(self.name, tag_name, line_ids, str(order)))
            self.command_processor.handle_command(production_command)

        elif request.state in ("READY", "SERVED", "DELIVERED", "INVALID"):
            production_command = ChangeProdStateCommand(request.order_id, self.get_view_name(), request.state)
            self.command_processor.handle_command(production_command)

    def _handle_refresh_view(self):
        threads = []
        for order in filter(lambda x: x.prod_state != ProdStates.INVALID, self.orders.values()):
            threads.append(self.view.handle_order(order, True))

        for thread in threads:
            thread.join()

        if isinstance(self.view, KdsView):
            self.view.refresh_end()

    def can_tag(self, order):
        return order.state in self.can_tag_states

    def can_change_prod_state(self, order):
        return order.state in self.can_change_prod_state_states

    def get_view_name(self):
        if isinstance(self.view, KdsView):
            return self.view.view_name
        return None

    @staticmethod
    def get_tag_by_state(state):
        if state == "ITEMS_DONE":
            return "done"
        elif state == "ITEMS_SERVED":
            return "served"
        return ""

    def get_order_line_ids(self, order, forbidden=None):
        # type: (ProductionOrder, Optional[List]) -> List[str]
        line_ids = []
        items = []
        for item in order.items:
            items.extend(item.joined_items)

        items.extend(order.items)

        for item in items:
            if not item.has_some_tag(ensure_iterable(forbidden)):
                self.get_item_line_ids(item, line_ids)

        return line_ids

    def get_item_line_ids(self, item, line_ids):
        if is_product(item):
            line_ids.append(item.get_line_id())
            for joined_item in item.joined_items:
                line_ids.append(joined_item.get_line_id())
        else:
            for son in item.items:
                self.get_item_line_ids(son, line_ids)

    def reload_view(self):
        if isinstance(self.view, KdsView):
            self.view.reload_handler()



class Config(object):

    def __init__(self):
        self.loader_path = None
        self.component_name = None
        self.required_services = None
        self.host = None
        self.port = None
        self.allowed_payment_methods = None
        self.available_sale_types = None
        self.ruptured_items_cache_time = None
        self.api_key_mapping = None
        self.nf_type = None
        self.cors_origin_hosts = None
        self.images_path = None
        self.totem_config_path = None
        self.up_sell_part_codes_path = None


class PosMap(object):

    def __init__(self, pos_id=None, api_key=None, abstract_pos_id=None, user_id=None, user_password=None):
        self.pos_id = pos_id
        self.api_key = api_key
        self.abstract_pos_id = abstract_pos_id
        self.user_id = user_id
        self.user_password = user_password



# -*- coding: utf-8 -*-

import collections
import json
from datetime import datetime
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from enum import Enum, IntEnum
from persistence import Connection

from helper import ExtendedEnum
from typing import List, Union, Optional


class Order(object):
    def __init__(self):
        self.order_id = None  # type: Union[int, None]
        self.pos_id = None  # type: Union[int, None]
        self.customer_cpf = None  # type: Union[str, None]
        self.discount_amount = None  # type: Union[float, None]
        self.sale_items = None  # type: Union[List[SaleItem], None]
        self.states = None  # type: Union[List[OrderState], None]
        self.tenders = None  # type: Union[List[OrderTender], None]
        self.tip = None  # type: Union[float, None]
        self.tip_rate = None  # type: Union[float, None]
        self.delivery_fee = None  # type: Union[float, None]
        self.discount_amount = None  # type: Union[float, None]
        self.sale_type = None  # type: Union[int, None]
        self.order_picture_xml = None
        self.total_gross = None  # type: Union[float, None]


class OrderKey(object):
    def __init__(self, pos_id, order_id):
        self.pos_id = pos_id
        self.order_id = order_id

    def __eq__(self, o):
        if isinstance(o, OrderKey):
            return o.order_id == self.order_id and o.pos_id == self.pos_id

        return False


class OrderParser(object):
    def __init__(self):
        return

    def parse_order(self, order_xml):
        # type: (ElementTree) -> Order

        pos_id = order_xml.get("posId") or self._get_pos_id_from_session_id(order_xml.get("sessionId"))
        order_id = int(order_xml.get("orderId"))
        flat_list = self.generate_flat_list(order_xml)
        cpf_element = order_xml.find('.//OrderProperty[@key="CUSTOMER_FISCAL_DOCUMENT"]')
        tip = float(order_xml.attrib['tip']) if 'tip' in order_xml.attrib else None
        order = self.generate_order(flat_list, order_id, cpf_element, str(pos_id), tip)
        order.discount_amount = float(order_xml.get('discountAmount') or 0.0)
        order.states = self.parse_order_states(order_xml.findall("StateHistory/State"))
        order.tenders = self.parse_tender(order_xml.findall("TenderHistory/Tender"))
        order.sale_type = int(order_xml.get("saleType"))
        order.order_picture_xml = order_xml
        order.tip_rate = order_xml.find('.//OrderProperty[@key="TIP_RATE"]')
        order.total_gross = float(order_xml.get('totalGross'))

        return order

    def generate_flat_list(self, order_xml):
        # type: (ElementTree) ->List[SaleItem]
        sale_items = self.parse_sale_lines(order_xml.findall("SaleLine"))
        sale_trees = self._convert_sale_lines_to_tree(sale_items)
        for sale_tree in sale_trees:
            self._fix_quantity_and_prices(sale_tree)
        flat_list = []  # type: List[SaleItem]
        for sale_tree in sale_trees:
            self._flat_tree_removing_non_priced_items(sale_tree, flat_list)
        return flat_list

    @staticmethod
    def parse_tender(tenders):
        # type: (ElementTree) ->  List[OrderTender]

        tender_list = []  # type: List[OrderTender]
        for tender in tenders:
            amount = float(tender.get("tenderAmount"))
            change = float(tender.get("change") or 0)
            tender_type = int(tender.get("tenderType"))
            tender_detail = tender.get("tenderDetail")
            tender_time = datetime.strptime(tender.get("timestamp"), "%Y-%m-%dT%H:%M:%S.%f")
            cnpj_auth = None
            bandeira = None
            auth_code = None
            accreditor_sat = None
            nfce_brand = None
            original_nfce_brand=None

            if tender_detail:
                tender_detail = json.loads(tender_detail)
                cnpj_auth = tender_detail['CNPJAuth'] if 'CNPJAuth' in tender_detail else None
                bandeira = tender_detail['Bandeira'] if 'Bandeira' in tender_detail else None
                auth_code = tender_detail['AuthCode'] if 'AuthCode' in tender_detail else None
                accreditor_sat = tender_detail.get('AccreditorSat', None)
                nfce_brand = tender_detail.get('NfceBrand', None)
                original_nfce_brand = tender_detail.get('OriginalNfceBrand', None)

            tender_desc = tender.get("tenderDescr")
            if amount > 0:
                order_tender = OrderTender(
                    amount=amount,
                    change=change,
                    tender_type=tender_type,
                    descr=tender_desc,
                    time=tender_time,
                    cnpj_auth=cnpj_auth,
                    bandeira=bandeira,
                    auth_code=auth_code,
                    accreditor_sat=accreditor_sat,
                    nfce_brand=nfce_brand,
                    original_nfce_brand=original_nfce_brand
                )
                tender_list.append(order_tender)
        return tender_list

    @staticmethod
    def parse_order_states(states):
        # type: (ElementTree) ->  List[OrderState]
        state_list = []  # type: List[OrderState]
        for state in states:
            timestamp = datetime.strptime(state.get("timestamp"), "%Y-%m-%dT%H:%M:%S.%f")
            timestamp_gmt = datetime.strptime(state.get("timestampGMT"), "%Y-%m-%dT%H:%M:%S.%f")
            order_state = OrderState(state.get("stateId"), state.get("state"), timestamp, timestamp_gmt)
            state_list.append(order_state)
        return state_list

    @staticmethod
    def generate_order(flat_list, order_id, cpf_element, pos_id, tip, delivery_fee=0.0, discount_amount=0.0):
        # type: (List[SaleItem], int, ElementTree, unicode, float, float, float) -> Order
        order = Order()
        order.order_id = order_id
        order.pos_id = pos_id
        order.sale_items = flat_list
        order.customer_cpf = cpf_element.get('value') if cpf_element is not None else None
        order.tip = tip
        order.delivery_fee = delivery_fee
        order.discount_amount = discount_amount
        return order

    def parse_sale_lines(self, order_sale_lines):
        sale_items = []  # type: List[SaleItem]
        for order_sale_line in order_sale_lines:
            sale_item = self.create_sale_item(order_sale_line)
            sale_item.type = self.get_sale_item_type(order_sale_line)

            sale_item.multiplied_quantity = float(order_sale_line.get("multipliedQty") or order_sale_line.get("qty") or 0)
            if order_sale_line.get("defaultQty") is not None:
                sale_item.default_quantity = int(order_sale_line.get("defaultQty"))

            sale_item.item_discount = self.get_sale_item_item_discount(order_sale_line)
            sale_item.taxes = self.get_sale_item_taxes(order_sale_line)
            sale_item.measure_unit = order_sale_line.get("measureUnit") or "UN"
            sale_items.append(sale_item)
        return sale_items

    @staticmethod
    def create_sale_item(order_sale_line):
        #  type: (ElementTree) -> SaleItem
        sale_item = SaleItem(
            int(order_sale_line.get("lineNumber")),
            int(order_sale_line.get("level")),
            int(order_sale_line.get("partCode")),
            order_sale_line.get("productName"),
            order_sale_line.get("itemId"),
            float(order_sale_line.get("itemPrice")) if order_sale_line.get("itemPrice") is not None else 0.0,
            float(order_sale_line.get("qty") or 0),
            float(order_sale_line.get("unitPrice")) if order_sale_line.get("unitPrice") is not None else 0.0,
            float(order_sale_line.get("addedUnitPrice")) if order_sale_line.get("addedUnitPrice") is not None else 0.0)

        if sale_item.quantity > 0:
            if sale_item.item_price > 0 and sale_item.unit_price == 0 and sale_item.added_unit_price == 0:
                sale_item.unit_price = sale_item.item_price / sale_item.quantity

            if (sale_item.unit_price > 0 or sale_item.added_unit_price > 0) and sale_item.item_price == 0:
                sale_item.unit_price = 0
                sale_item.added_unit_price = 0

        return sale_item

    def get_sale_item_taxes(self, order_sale_line):
        #  type: (ElementTree) -> List[TaxItem]
        taxes = []  # type: (List[TaxItem])
        for tax_item_xml in order_sale_line.findall("TaxItem"):
            tax_item = self._parse_tax_item(tax_item_xml)
            taxes.append(tax_item)
        return taxes

    @staticmethod
    def get_sale_item_item_discount(order_sale_line):
        #  type: (ElementTree) -> float
        if order_sale_line.get("itemDiscount") is not None:
            return float(order_sale_line.get("itemDiscount"))
        return 0.0

    @staticmethod
    def get_sale_item_type(order_sale_line):
        #  type: (ElementTree) -> unicode
        if order_sale_line.get("itemType") == "COMBO":
            return SaleItemType.combo
        elif order_sale_line.get("itemType") == "OPTION":
            return SaleItemType.option
        else:
            return SaleItemType.product

    @staticmethod
    def _parse_tax_item(tax_item_xml):
        # type: (Element) -> TaxItem
        tax_item = TaxItem()
        tax_index = tax_item_xml.get("taxIndex")
        tax_item.tax_name = tax_index.split(";")[0]
        tax_item.tax_index = tax_index.split(";")[1]
        tax_item.tax_rule_id = int(tax_item_xml.get("taxRuleId"))
        tax_item.tax_rate = float(tax_item_xml.get("taxRate"))
        if tax_item_xml.get("taxIncluded") is not None:
            tax_item.tax_included = int(tax_item_xml.get("taxIncluded"))
        tax_item.base_amount_bd = float(tax_item_xml.get("baseAmountBD"))
        tax_item.base_amount_ad = float(tax_item_xml.get("baseAmountAD"))
        tax_item.tax_amount_bd = float(tax_item_xml.get("taxAmountBD"))
        tax_item.tax_amount_ad = float(tax_item_xml.get("taxAmountAD"))

        return tax_item

    def _fix_quantity_and_prices(self, sale_tree, parent=None, grand_parent=None):
        # type: (SaleItem, SaleItem, SaleItem) -> None
        sale_tree = self._fix_quantity_by_ancestors(sale_tree, parent, grand_parent)

        sale_tree.unit_discount = (sale_tree.item_discount / sale_tree.quantity) if sale_tree.quantity > 0 else 0.0
        sale_tree.item_discount = sale_tree.item_discount if sale_tree.quantity > 0 else 0.0

        for tax in sale_tree.taxes:
            tax *= sale_tree.fix_qty_rate

        for son in sale_tree.sons:
            self._fix_quantity_and_prices(son, sale_tree, parent)

    def _fix_quantity_by_ancestors(self, sale_tree, parent, grand_parent):
        # type: (SaleItem, SaleItem, SaleItem) -> SaleItem
        sale_tree = self._verify_product_inside_option_inside_combo(sale_tree, parent, grand_parent)
        sale_tree = self._verify_product_inside_combo(sale_tree, parent)
        sale_tree = self._verify_product_inside_option_inside_product(sale_tree, parent, grand_parent)
        return sale_tree

    @staticmethod
    def _verify_product_inside_option_inside_combo(sale_tree, parent, grand_parent):
        # type: (SaleItem, SaleItem, SaleItem) -> SaleItem
        if sale_tree.type != SaleItemType.product:
            return sale_tree

        if parent is None or parent.type != SaleItemType.option:
            return sale_tree

        if grand_parent is None or grand_parent.type != SaleItemType.combo:
            return sale_tree

        sale_tree.quantity = sale_tree.quantity * grand_parent.multiplied_quantity
        if sale_tree.quantity > 0 and sale_tree.item_price > 0:
            sale_tree.fix_qty_rate = sale_tree.quantity * sale_tree.unit_price / sale_tree.item_price

        return sale_tree

    @staticmethod
    def _verify_product_inside_combo(sale_tree, parent):
        # type: (SaleItem, SaleItem) -> SaleItem
        if sale_tree.type != SaleItemType.product:
            return sale_tree

        if parent is None or parent.type != SaleItemType.combo:
            return sale_tree

        sale_tree.quantity = sale_tree.quantity * parent.multiplied_quantity
        if sale_tree.quantity > 0 and sale_tree.item_price > 0:
            sale_tree.fix_qty_rate = sale_tree.quantity * sale_tree.unit_price / sale_tree.item_price

        return sale_tree

    @staticmethod
    def _verify_product_inside_option_inside_product(sale_tree, parent, grand_parent):
        # type: (SaleItem, SaleItem, SaleItem) -> SaleItem
        if sale_tree.type != SaleItemType.product:
            return sale_tree

        if parent is None or parent.type != SaleItemType.option:
            return sale_tree

        if grand_parent is None or grand_parent.type != SaleItemType.product:
            return sale_tree

        sale_tree.quantity = (sale_tree.quantity or 0) - (sale_tree.default_quantity or 0)
        sale_tree.quantity *= grand_parent.multiplied_quantity

        sale_tree.unit_price = sale_tree.added_unit_price if sale_tree.added_unit_price > 0 else sale_tree.unit_price
        if sale_tree.quantity > 0 and sale_tree.item_price > 0:
            sale_tree.fix_qty_rate = sale_tree.quantity * sale_tree.unit_price / sale_tree.item_price

        return sale_tree

    def _build_item_tree(self, sale_item):
        # type: (SaleItem) -> None
        all_sons = []  # type: List[SaleItem]
        for son in sale_item.sons:
            all_sons.append(son)

        # Adding direct sons from actual item
        current_context = sale_item.item_id + "." + str(sale_item.part_code)
        sale_item.sons = []
        sons_to_remove = []
        self.add_sons(all_sons, current_context, sale_item, sons_to_remove)

        # Removing sons already added from full list
        self.remove_from_list(all_sons, sons_to_remove)

        sons_to_remove = []
        self.add_children(all_sons, sale_item, sons_to_remove)

        if len(sons_to_remove) != len(all_sons):
            # TODO: FIX "1.7100018.7300000.6019.10005000" partCode="8700017" (OPTION 10005000 NOT IN ORDERPICT)
            # raise Exception("Sons not added to any parent")
            pass

        for son in sale_item.sons:
            self._build_item_tree(son)

    @staticmethod
    def add_children(all_sons, sale_item, sons_to_remove):
        for current_son in sale_item.sons:
            current_son_context = current_son.item_id + "." + str(current_son.part_code)
            for son in all_sons:
                if son.item_id.startswith(current_son_context):
                    current_son.sons.append(son)
                    sons_to_remove.append(son)

    @staticmethod
    def remove_from_list(all_sons, sons_to_remove):
        for son in sons_to_remove:
            all_sons.remove(son)

    @staticmethod
    def add_sons(all_sons, current_context, sale_item, sons_to_remove):
        for son in all_sons:
            if son.item_id == current_context:
                sale_item.sons.append(son)
                son.parent = sale_item
                sons_to_remove.append(son)

    def _flat_tree_removing_non_priced_items(self, sale_tree, flat_list):
        # type: (SaleItem, List[SaleItem]) -> None
        if (sale_tree.unit_price or sale_tree.added_unit_price > 0) and sale_tree.quantity > 0:
            flat_list.append(sale_tree)

        for son in sale_tree.sons:
            self._flat_tree_removing_non_priced_items(son, flat_list)

    def _convert_sale_lines_to_tree(self, order_sale_lines):
        # type: (List[SaleItem]) -> List[SaleItem]

        # Find all items with context 1
        level_0_items = filter(lambda x: x.level == 0, order_sale_lines)

        # Insert items into correspondent level 0 lines
        for level_0_item in level_0_items:
            for order_sale_line in order_sale_lines:
                if order_sale_line.line_number != level_0_item.line_number:
                    continue

                if level_0_item != order_sale_line:
                    level_0_item.sons.append(order_sale_line)
                    order_sale_line.parent = level_0_item

        for item in level_0_items:
            self._build_item_tree(item)

        return sorted(level_0_items, cmp=lambda item1, item2: item1.line_number - item2.line_number)

    @staticmethod
    def _find_level_0_items(sale_items):
        # type: (List[SaleItem]) -> List[SaleItem]
        return filter(lambda x: x.level == 0, sale_items)

    @staticmethod
    def _add_same_line_number(sale_items, parent_items):
        # type: (List[SaleItem], List[SaleItem]) -> None
        for parent_item in parent_items:
            for sale_item in sale_items:
                if sale_item.line_number == parent_item.line_number and sale_item.part_code != parent_item.part_code:
                    parent_item.sons.append(sale_item)
                    sale_item.parent = parent_item

    @staticmethod
    def _add_direct_sons(all_sons, sale_item):
        # type: (List[SaleItem], SaleItem) -> None
        current_context = sale_item.item_id + "." + str(sale_item.part_code)
        sons_to_remove = []
        for son in all_sons:
            if son.item_id == current_context:
                sale_item.sons.append(son)
                son.parent = sale_item
                sons_to_remove.append(son)

        for son in sons_to_remove:
            all_sons.remove(son)

    @staticmethod
    def _add_descendants(all_sons, sale_item):
        # type: (List[SaleItem], SaleItem) -> None
        sons_to_remove = []

        current_son_context = sale_item.item_id + "." + str(sale_item.part_code)
        for son in all_sons:
            if son.item_id.startswith(current_son_context):
                sale_item.sons.append(son)
                sons_to_remove.append(son)

        for son_to_remove in sons_to_remove:
            all_sons.remove(son_to_remove)

    @staticmethod
    def _get_pos_id_from_session_id(session_id):
        # type: (str) -> int
        try:
            pos_id_start_index = session_id.index("pos=")
            pos_id_end_index = session_id.index(",", pos_id_start_index)
            pos_id = session_id[pos_id_start_index + 4:pos_id_end_index]
        except ValueError:
            pos_id = "0"

        return int(pos_id)


class OrderState(object):
    def __init__(self, key, name, timestamp, timestamp_gmt):
        self.key = key  # type: int
        self.name = name  # type: unicode
        self.timestamp = timestamp  # type: datetime
        self.timestamp_gmt = timestamp_gmt  # type: datetime

    def __eq__(self, o):
        if isinstance(o, OrderState):
            return o.key == self.key \
                   and o.name == self.name \
                   and o.timestamp == self.timestamp \
                   and o.timestamp_gmt == self.timestamp_gmt

        return False


class OrderTender(object):
    def __init__(
        self,
        amount,                     # type: float
        change,                     # type: float
        tender_type,                # type: int
        descr,                      # type: unicode
        time,                       # type: datetime
        cnpj_auth,                  # type: unicode
        bandeira,                   # type: unicode
        auth_code,                  # type: Optional[unicode]
        accreditor_sat=None,        # type: Optional[unicode]
        nfce_brand=None,            # type: Optional[unicode]
        original_nfce_brand=None    # type: Optional[unicode]
    ):
        # type: (...) -> None

        self.amount = amount
        self.change = change
        self.type = tender_type
        self.descr = descr
        self.time = time
        self.cnpj_auth = cnpj_auth
        self.bandeira = bandeira
        self.auth_code = auth_code
        self.accreditor_sat = accreditor_sat
        self.nfce_brand = nfce_brand
        self.original_nfce_brand = original_nfce_brand

    def __eq__(self, o):
        if isinstance(o, OrderTender):
            return o.amount == self.amount \
                   and o.change == self.change \
                   and o.type == self.type \
                   and o.descr == self.descr \
                   and o.time == self.time \
                   and o.cnpj_auth == self.cnpj_auth \
                   and o.bandeira == self.bandeira \
                   and o.auth_code == self.auth_code \
                   and o.accreditor_sat == self.accreditor_sat \
                   and o.nfce_brand == self.nfce_brand \
                   and o.original_nfce_brand == self.original_nfce_brand

        return False


class PosConnection(object):
    def __init__(self, pos_id, connection):
        # type: (str, Connection) -> None
        self.pos_id = pos_id
        self.connection = connection

    def __del__(self):
        return self.connection.__del__()

    def close(self):
        return self.connection.close()

    def is_multi_instance(self):
        return self.connection.is_multi_instance()

    def set_dbname(self, db_name):
        # type: (str) -> None
        return self.connection.set_dbname(db_name)

    def query(self, sql):
        return self.connection.query(sql)

    def pquery(self, procedure_name, **kargs):
        return self.connection.pquery(procedure_name, **kargs)

    def select(self, sql):
        return self.connection.select(sql)

    def pselect(self, procedure_name, **kargs):
        return self.connection.pselect(procedure_name, **kargs)

    def escape(self, string):
        return self.connection.escape(string)

    def transaction_start(self):
        return self.connection.transaction_start()

    def transaction_end(self):
        return self.connection.transaction_end()

    def transaction_mode_get(self):
        return self.connection.transaction_mode_get()

    def transaction_mode_set(self, mode):
        return self.connection.transaction_mode_set(mode)


class SaleItemType(object):
    combo = "Combo"
    option = "Option"
    product = "Product"


class SaleItem(object):
    def __init__(self, line_number=None, level=None, part_code=None, product_name=None, item_id=None, item_price=None,
                 quantity=None, unit_price=None, added_unit_price=0.0, item_discount=0.0, unit_discount=0.0,
                 item_additional=0.0, unit_additional=0.0, measure_unit="UN"):
        # type: (int, int, int, unicode, Optional[unicode], float, float, float, float, float, Optional[str]) -> None
        self.line_number = line_number
        self.level = level
        self.part_code = part_code
        self.product_name = product_name  # type: unicode
        self.item_id = item_id
        self.item_price = item_price
        self.quantity = quantity
        self.unit_price = unit_price  # type: float
        self.added_unit_price = added_unit_price  # type: float
        self.item_discount = item_discount
        self.unit_discount = unit_discount
        self.item_additional = item_additional
        self.unit_additional = unit_additional
        self.measure_unit = measure_unit

        self.fix_qty_rate = 1

        self.default_quantity = None  # type: None
        self.type = None  # type: unicode
        self.multiplied_quantity = None  # type: int

        self.parent = None  # type: SaleItem
        self.sons = []  # type: List[SaleItem]
        self.taxes = []  # type: List[TaxItem]

    def get_tax_item(self, tax_name):
        # type: (unicode) -> Optional[TaxItem]
        for tax_item in self.taxes:
            if tax_item.tax_name == tax_name:
                unit_price = self.unit_price if self.unit_price > 0 else self.added_unit_price
                if tax_item.base_amount_bd == 0 or unit_price == 0:
                    return None

                multiplier = round(self.quantity / (tax_item.base_amount_bd / unit_price))
                return tax_item * multiplier

        return None

    def __str__(self):
        return ("lineNumber: " + str(self.line_number) + ", " +
                "level: " + str(self.level) + ", " +
                "part_code: " + str(self.part_code) + ", " +
                "item_id: " + self.item_id + ", " +
                "quantity: " + str(self.quantity) + ", " +
                "unit_price: " + str(self.unit_price) + ", " +
                "item_price: " + str(self.item_price) + ", " +
                "taxes: " + str(self.taxes) + ", " +
                "sons: " + str(self.sons) + ", " +
                "parent: " + str(self.parent) + ", ")

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, SaleItem):
            return self.level == other.level and self.item_id == other.item_id and self.part_code == other.part_code

        return False

    def __hash__(self):
        return hash((self.level, self.item_id, self.part_code))


class Tax(object):
    def __init__(self, code, name, rate, fiscal_index, tax_processor, params):
        # type: (unicode, unicode, float, unicode, unicode) -> None
        self.code = code
        self.name = name
        self.rate = rate
        self.fiscal_index = fiscal_index
        self.tax_processor = tax_processor
        self.params = params


class TaxItem(object):
    def __init__(self):
        self.tax_rule_id = None  # type: unicode
        self.tax_index = None  # type: unicode
        self.tax_name = None  # type: unicode
        self.tax_code = None  # type: unicode
        self.base_amount_bd = None  # type: float
        self.base_amount_ad = None  # type: float
        self.tax_amount_bd = None  # type: float
        self.tax_amount_ad = None  # type: float
        self.tax_included = None  # type: float
        self.tax_rate = None  # type: float

    def __str__(self):
        return ("taxRuleId: " + str(self.tax_rule_id) + ", " +
                "taxIndex: " + str(self.tax_index) + ", " +
                "taxName: " + str(self.tax_name) + ", " +
                "taxRate: " + str(self.tax_rate) + ", " +
                "baseAmountBd: " + str(self.base_amount_bd) + ", " +
                "baseAmountAd: " + str(self.base_amount_ad) + ", " +
                "taxAmountBd: " + str(self.tax_amount_bd) + ", " +
                "taxAmountAd: " + str(self.tax_amount_ad) + ", " +
                "taxIncluded: " + str(self.tax_included))

    def __repr__(self):
        return self.__str__()

    def __mul__(self, o):
        self.base_amount_ad *= o
        self.base_amount_bd *= o
        self.tax_amount_ad *= o
        self.tax_amount_bd *= o

        return self


class TenderType(object):
    cash = 0
    credit = 1
    debit = 2
    voucher = 3
    digital_wallet = 4
    connection_test = 999


class TenderTypeDescription(object):
    NON_ELECTRONIC = "NON_ELECTRONIC"
    CREDIT = "CREDIT_CARD"
    DEBIT = "DEBIT_CARD"
    VOUCHER = "VOUCHER_CARD"
    DIGITAL_WALLET = "DIGITAL_WALLET"


class SaleType(object):
    EAT_IN = 0
    TAKE_OUT = 1
    DRIVE_THRU = 2
    DELIVERY = 3
    ONLINE = 4
    PICKUP = 5
    PHONE = 6
    CATERING = 7


class XmlTenderType(object):
    cash = 1
    credit = 3
    debit = 4
    others = 99


class XmlTenderTypeFromPOSMapper(object):
    @staticmethod
    def get(tender_type):
        # type: (int) -> int
        if tender_type == TenderType.cash:
            return XmlTenderType.cash
        if tender_type == TenderType.credit:
            return XmlTenderType.credit
        if tender_type == TenderType.debit:
            return XmlTenderType.debit
        return XmlTenderType.others


class POSTenderTypeFromXmlMapper(object):
    @staticmethod
    def get(tender_type):
        # type: (int) -> Union[int, None]
        if tender_type == XmlTenderType.cash:
            return TenderType.cash
        if tender_type == XmlTenderType.credit:
            return TenderType.credit
        if tender_type == XmlTenderType.debit:
            return TenderType.debit
        return XmlTenderType.others


class DeliveryTenderTypes(Enum):
    SPOON_ROCKET = 27
    IFOOD = 28
    LOGGI = 29
    APP = 30
    UBER_EATS = 31
    GLOVO = 32

    @staticmethod
    def list_tender_values():
        delivery_tender_types = DeliveryTenderTypes  # type: collections.Iterable
        return [x.value for x in delivery_tender_types]


class TableStatus(IntEnum):
    UNMAPPED_STATUS = 0
    AVAILABLE = 1
    WAITING_TO_BE_SEATED = 2
    SEATED = 3
    IN_PROGRESS = 4
    LINKED = 5
    TOTALED = 6
    CLOSED = 7

    @staticmethod
    def list_values():
        return [x.value for x in TableStatus]

    @staticmethod
    def get_status_label(status):
        # type: (int) -> Union[str, None]
        if status not in [x for x in TableStatus.list_values()]:
            return TableStatus.Unmaped.name

        for table_status in TableStatus:
            if status == table_status.value:
                return table_status.name


class TableStatusDescr(object):
    AVAILABLE = "Available"
    WAITING_TO_BE_SEATED = "Waiting2BSeated"
    SEATED = "Seated"
    IN_PROGRESS = "InProgress"
    LINKED = "Linked"
    TOTALED = "Totaled"
    CLOSED = "Closed"


class POSType(object):
    DT = "DT"
    DL = "DL"
    TT = "TT"
    FC = "FC"
    KT = "KT"


class SaleType(object):
    EAT_IN = "EAT_IN"
    TAKE_OUT = "TAKE_OUT"
    DRIVE_THRU = "DRIVE_THRU"
    DELIVERY = "DELIVERY"
    ONLINE = "ONLINE"
    PICKUP = "PICKUP"
    PHONE = "PHONE"
    CATERING = "CATERING"


class PODType(object):
    FC = "FC"
    TS = "TS"
    FL = "FL"
    TT = "TT"
    KT = "KT"


class POSTypeMapper(object):

    @staticmethod
    def get(sale_type, pod_type):
        # type: (str, str) -> Optional[str]

        if sale_type == SaleType.DRIVE_THRU:
            return POSType.DT
        if sale_type == SaleType.DELIVERY:
            return POSType.DL
        if pod_type == PODType.TT:
            return POSType.TT
        if sale_type in [SaleType.EAT_IN, SaleType.TAKE_OUT]:
            return POSType.FC

        return None

    @staticmethod
    def get_types():
        # type: () -> List[str]

        return [POSType.DL, POSType.DT, POSType.FC, POSType.TT]

    @staticmethod
    def get_kitchen_type():
        # type: () -> str

        return POSType.KT


class TipType(object):
    PERCENTAGE = "Percentage"
    AMOUNT = "Amount"


class OrderStatus(ExtendedEnum):
    UNDEFINED = 0
    IN_PROGRESS = 1
    STORED = 2
    TOTALED = 3
    VOIDED = 4
    PAID = 5
    RECALLED = 6
    SYSTEM_VOIDED = 7
    ABANDONED = 8



import json

from bustoken import TK_REMOTE_ORDER_GET_RUPTURED_ITEMS
from msgbus import FM_STRING, TK_SYS_NAK
from sysactions import send_message
from systools import sys_log_exception


def get_ruptured_products_from_remote_order():
    ruptured_items = []
    try:
        msg = send_message("RemoteOrder", TK_REMOTE_ORDER_GET_RUPTURED_ITEMS, FM_STRING, "", timeout=60000000)
        if msg.token == TK_SYS_NAK:
            raise Exception()

        ruptured_items = json.loads(msg.data)
    except Exception as _:
        sys_log_exception("Error getting not ruptured items")

    return ruptured_items



# encoding: utf-8
import json

from benefitsreport.repository import BenefitsOrderRepository
from reportsdata.benefitsreport import BenefitsGenerator, BenefitsPresenter


class ReportGenerator(object):
    @staticmethod
    def generate(mb_context, _, *args):

        initial_date = args[4]
        end_date = args[5]
        operator_id = None if args[6] == '0' else args[6]
        benefits = json.loads(args[8])
        pos_list = json.loads(args[len(args) - 1])
        order_repository = BenefitsOrderRepository(mb_context, pos_list)
        generator = BenefitsGenerator(order_repository, initial_date, end_date, operator_id, benefits)
        presenter = BenefitsPresenter(38)

        return generator, presenter




# -*- coding: utf-8 -*-
"""The runtime functions and state used by compiled templates."""
import sys
from itertools import chain
from types import MethodType

from markupsafe import escape  # noqa: F401
from markupsafe import Markup
from markupsafe import soft_unicode

from ._compat import abc
from ._compat import imap
from ._compat import implements_iterator
from ._compat import implements_to_string
from ._compat import iteritems
from ._compat import PY2
from ._compat import string_types
from ._compat import text_type
from ._compat import with_metaclass
from .exceptions import TemplateNotFound  # noqa: F401
from .exceptions import TemplateRuntimeError  # noqa: F401
from .exceptions import UndefinedError
from .nodes import EvalContext
from .utils import concat
from .utils import evalcontextfunction
from .utils import internalcode
from .utils import missing
from .utils import Namespace  # noqa: F401
from .utils import object_type_repr

# these variables are exported to the template runtime
exported = [
    "LoopContext",
    "TemplateReference",
    "Macro",
    "Markup",
    "TemplateRuntimeError",
    "missing",
    "concat",
    "escape",
    "markup_join",
    "unicode_join",
    "to_string",
    "identity",
    "TemplateNotFound",
    "Namespace",
    "Undefined",
]

#: the name of the function that is used to convert something into
#: a string.  We can just use the text type here.
to_string = text_type


def identity(x):
    """Returns its argument. Useful for certain things in the
    environment.
    """
    return x


def markup_join(seq):
    """Concatenation that escapes if necessary and converts to unicode."""
    buf = []
    iterator = imap(soft_unicode, seq)
    for arg in iterator:
        buf.append(arg)
        if hasattr(arg, "__html__"):
            return Markup(u"").join(chain(buf, iterator))
    return concat(buf)


def unicode_join(seq):
    """Simple args to unicode conversion and concatenation."""
    return concat(imap(text_type, seq))


def new_context(
    environment,
    template_name,
    blocks,
    vars=None,
    shared=None,
    globals=None,
    locals=None,
):
    """Internal helper for context creation."""
    if vars is None:
        vars = {}
    if shared:
        parent = vars
    else:
        parent = dict(globals or (), **vars)
    if locals:
        # if the parent is shared a copy should be created because
        # we don't want to modify the dict passed
        if shared:
            parent = dict(parent)
        for key, value in iteritems(locals):
            if value is not missing:
                parent[key] = value
    return environment.context_class(environment, parent, template_name, blocks)


class TemplateReference(object):
    """The `self` in templates."""

    def __init__(self, context):
        self.__context = context

    def __getitem__(self, name):
        blocks = self.__context.blocks[name]
        return BlockReference(name, self.__context, blocks, 0)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.__context.name)


def _get_func(x):
    return getattr(x, "__func__", x)


class ContextMeta(type):
    def __new__(mcs, name, bases, d):
        rv = type.__new__(mcs, name, bases, d)
        if bases == ():
            return rv

        resolve = _get_func(rv.resolve)
        default_resolve = _get_func(Context.resolve)
        resolve_or_missing = _get_func(rv.resolve_or_missing)
        default_resolve_or_missing = _get_func(Context.resolve_or_missing)

        # If we have a changed resolve but no changed default or missing
        # resolve we invert the call logic.
        if (
            resolve is not default_resolve
            and resolve_or_missing is default_resolve_or_missing
        ):
            rv._legacy_resolve_mode = True
        elif (
            resolve is default_resolve
            and resolve_or_missing is default_resolve_or_missing
        ):
            rv._fast_resolve_mode = True

        return rv


def resolve_or_missing(context, key, missing=missing):
    if key in context.vars:
        return context.vars[key]
    if key in context.parent:
        return context.parent[key]
    return missing


class Context(with_metaclass(ContextMeta)):
    """The template context holds the variables of a template.  It stores the
    values passed to the template and also the names the template exports.
    Creating instances is neither supported nor useful as it's created
    automatically at various stages of the template evaluation and should not
    be created by hand.

    The context is immutable.  Modifications on :attr:`parent` **must not**
    happen and modifications on :attr:`vars` are allowed from generated
    template code only.  Template filters and global functions marked as
    :func:`contextfunction`\\s get the active context passed as first argument
    and are allowed to access the context read-only.

    The template context supports read only dict operations (`get`,
    `keys`, `values`, `items`, `iterkeys`, `itervalues`, `iteritems`,
    `__getitem__`, `__contains__`).  Additionally there is a :meth:`resolve`
    method that doesn't fail with a `KeyError` but returns an
    :class:`Undefined` object for missing variables.
    """

    # XXX: we want to eventually make this be a deprecation warning and
    # remove it.
    _legacy_resolve_mode = False
    _fast_resolve_mode = False

    def __init__(self, environment, parent, name, blocks):
        self.parent = parent
        self.vars = {}
        self.environment = environment
        self.eval_ctx = EvalContext(self.environment, name)
        self.exported_vars = set()
        self.name = name

        # create the initial mapping of blocks.  Whenever template inheritance
        # takes place the runtime will update this mapping with the new blocks
        # from the template.
        self.blocks = dict((k, [v]) for k, v in iteritems(blocks))

        # In case we detect the fast resolve mode we can set up an alias
        # here that bypasses the legacy code logic.
        if self._fast_resolve_mode:
            self.resolve_or_missing = MethodType(resolve_or_missing, self)

    def super(self, name, current):
        """Render a parent block."""
        try:
            blocks = self.blocks[name]
            index = blocks.index(current) + 1
            blocks[index]
        except LookupError:
            return self.environment.undefined(
                "there is no parent block called %r." % name, name="super"
            )
        return BlockReference(name, self, blocks, index)

    def get(self, key, default=None):
        """Returns an item from the template context, if it doesn't exist
        `default` is returned.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def resolve(self, key):
        """Looks up a variable like `__getitem__` or `get` but returns an
        :class:`Undefined` object with the name of the name looked up.
        """
        if self._legacy_resolve_mode:
            rv = resolve_or_missing(self, key)
        else:
            rv = self.resolve_or_missing(key)
        if rv is missing:
            return self.environment.undefined(name=key)
        return rv

    def resolve_or_missing(self, key):
        """Resolves a variable like :meth:`resolve` but returns the
        special `missing` value if it cannot be found.
        """
        if self._legacy_resolve_mode:
            rv = self.resolve(key)
            if isinstance(rv, Undefined):
                rv = missing
            return rv
        return resolve_or_missing(self, key)

    def get_exported(self):
        """Get a new dict with the exported variables."""
        return dict((k, self.vars[k]) for k in self.exported_vars)

    def get_all(self):
        """Return the complete context as dict including the exported
        variables.  For optimizations reasons this might not return an
        actual copy so be careful with using it.
        """
        if not self.vars:
            return self.parent
        if not self.parent:
            return self.vars
        return dict(self.parent, **self.vars)

    @internalcode
    def call(__self, __obj, *args, **kwargs):  # noqa: B902
        """Call the callable with the arguments and keyword arguments
        provided but inject the active context or environment as first
        argument if the callable is a :func:`contextfunction` or
        :func:`environmentfunction`.
        """
        if __debug__:
            __traceback_hide__ = True  # noqa

        # Allow callable classes to take a context
        if hasattr(__obj, "__call__"):  # noqa: B004
            fn = __obj.__call__
            for fn_type in (
                "contextfunction",
                "evalcontextfunction",
                "environmentfunction",
            ):
                if hasattr(fn, fn_type):
                    __obj = fn
                    break

        if callable(__obj):
            if getattr(__obj, "contextfunction", False) is True:
                args = (__self,) + args
            elif getattr(__obj, "evalcontextfunction", False) is True:
                args = (__self.eval_ctx,) + args
            elif getattr(__obj, "environmentfunction", False) is True:
                args = (__self.environment,) + args
        try:
            return __obj(*args, **kwargs)
        except StopIteration:
            return __self.environment.undefined(
                "value was undefined because "
                "a callable raised a "
                "StopIteration exception"
            )

    def derived(self, locals=None):
        """Internal helper function to create a derived context.  This is
        used in situations where the system needs a new context in the same
        template that is independent.
        """
        context = new_context(
            self.environment, self.name, {}, self.get_all(), True, None, locals
        )
        context.eval_ctx = self.eval_ctx
        context.blocks.update((k, list(v)) for k, v in iteritems(self.blocks))
        return context

    def _all(meth):  # noqa: B902
        def proxy(self):
            return getattr(self.get_all(), meth)()

        proxy.__doc__ = getattr(dict, meth).__doc__
        proxy.__name__ = meth
        return proxy

    keys = _all("keys")
    values = _all("values")
    items = _all("items")

    # not available on python 3
    if PY2:
        iterkeys = _all("iterkeys")
        itervalues = _all("itervalues")
        iteritems = _all("iteritems")
    del _all

    def __contains__(self, name):
        return name in self.vars or name in self.parent

    def __getitem__(self, key):
        """Lookup a variable or raise `KeyError` if the variable is
        undefined.
        """
        item = self.resolve_or_missing(key)
        if item is missing:
            raise KeyError(key)
        return item

    def __repr__(self):
        return "<%s %s of %r>" % (
            self.__class__.__name__,
            repr(self.get_all()),
            self.name,
        )


abc.Mapping.register(Context)


class BlockReference(object):
    """One block on a template reference."""

    def __init__(self, name, context, stack, depth):
        self.name = name
        self._context = context
        self._stack = stack
        self._depth = depth

    @property
    def super(self):
        """Super the block."""
        if self._depth + 1 >= len(self._stack):
            return self._context.environment.undefined(
                "there is no parent block called %r." % self.name, name="super"
            )
        return BlockReference(self.name, self._context, self._stack, self._depth + 1)

    @internalcode
    def __call__(self):
        rv = concat(self._stack[self._depth](self._context))
        if self._context.eval_ctx.autoescape:
            rv = Markup(rv)
        return rv


@implements_iterator
class LoopContext:
    """A wrapper iterable for dynamic ``for`` loops, with information
    about the loop and iteration.
    """

    #: Current iteration of the loop, starting at 0.
    index0 = -1

    _length = None
    _after = missing
    _current = missing
    _before = missing
    _last_changed_value = missing

    def __init__(self, iterable, undefined, recurse=None, depth0=0):
        """
        :param iterable: Iterable to wrap.
        :param undefined: :class:`Undefined` class to use for next and
            previous items.
        :param recurse: The function to render the loop body when the
            loop is marked recursive.
        :param depth0: Incremented when looping recursively.
        """
        self._iterable = iterable
        self._iterator = self._to_iterator(iterable)
        self._undefined = undefined
        self._recurse = recurse
        #: How many levels deep a recursive loop currently is, starting at 0.
        self.depth0 = depth0

    @staticmethod
    def _to_iterator(iterable):
        return iter(iterable)

    @property
    def length(self):
        """Length of the iterable.

        If the iterable is a generator or otherwise does not have a
        size, it is eagerly evaluated to get a size.
        """
        if self._length is not None:
            return self._length

        try:
            self._length = len(self._iterable)
        except TypeError:
            iterable = list(self._iterator)
            self._iterator = self._to_iterator(iterable)
            self._length = len(iterable) + self.index + (self._after is not missing)

        return self._length

    def __len__(self):
        return self.length

    @property
    def depth(self):
        """How many levels deep a recursive loop currently is, starting at 1."""
        return self.depth0 + 1

    @property
    def index(self):
        """Current iteration of the loop, starting at 1."""
        return self.index0 + 1

    @property
    def revindex0(self):
        """Number of iterations from the end of the loop, ending at 0.

        Requires calculating :attr:`length`.
        """
        return self.length - self.index

    @property
    def revindex(self):
        """Number of iterations from the end of the loop, ending at 1.

        Requires calculating :attr:`length`.
        """
        return self.length - self.index0

    @property
    def first(self):
        """Whether this is the first iteration of the loop."""
        return self.index0 == 0

    def _peek_next(self):
        """Return the next element in the iterable, or :data:`missing`
        if the iterable is exhausted. Only peeks one item ahead, caching
        the result in :attr:`_last` for use in subsequent checks. The
        cache is reset when :meth:`__next__` is called.
        """
        if self._after is not missing:
            return self._after

        self._after = next(self._iterator, missing)
        return self._after

    @property
    def last(self):
        """Whether this is the last iteration of the loop.

        Causes the iterable to advance early. See
        :func:`itertools.groupby` for issues this can cause.
        The :func:`groupby` filter avoids that issue.
        """
        return self._peek_next() is missing

    @property
    def previtem(self):
        """The item in the previous iteration. Undefined during the
        first iteration.
        """
        if self.first:
            return self._undefined("there is no previous item")

        return self._before

    @property
    def nextitem(self):
        """The item in the next iteration. Undefined during the last
        iteration.

        Causes the iterable to advance early. See
        :func:`itertools.groupby` for issues this can cause.
        The :func:`groupby` filter avoids that issue.
        """
        rv = self._peek_next()

        if rv is missing:
            return self._undefined("there is no next item")

        return rv

    def cycle(self, *args):
        """Return a value from the given args, cycling through based on
        the current :attr:`index0`.

        :param args: One or more values to cycle through.
        """
        if not args:
            raise TypeError("no items for cycling given")

        return args[self.index0 % len(args)]

    def changed(self, *value):
        """Return ``True`` if previously called with a different value
        (including when called for the first time).

        :param value: One or more values to compare to the last call.
        """
        if self._last_changed_value != value:
            self._last_changed_value = value
            return True

        return False

    def __iter__(self):
        return self

    def __next__(self):
        if self._after is not missing:
            rv = self._after
            self._after = missing
        else:
            rv = next(self._iterator)

        self.index0 += 1
        self._before = self._current
        self._current = rv
        return rv, self

    @internalcode
    def __call__(self, iterable):
        """When iterating over nested data, render the body of the loop
        recursively with the given inner iterable data.

        The loop must have the ``recursive`` marker for this to work.
        """
        if self._recurse is None:
            raise TypeError(
                "The loop must have the 'recursive' marker to be called recursively."
            )

        return self._recurse(iterable, self._recurse, depth=self.depth)

    def __repr__(self):
        return "<%s %d/%d>" % (self.__class__.__name__, self.index, self.length)


class Macro(object):
    """Wraps a macro function."""

    def __init__(
        self,
        environment,
        func,
        name,
        arguments,
        catch_kwargs,
        catch_varargs,
        caller,
        default_autoescape=None,
    ):
        self._environment = environment
        self._func = func
        self._argument_count = len(arguments)
        self.name = name
        self.arguments = arguments
        self.catch_kwargs = catch_kwargs
        self.catch_varargs = catch_varargs
        self.caller = caller
        self.explicit_caller = "caller" in arguments
        if default_autoescape is None:
            default_autoescape = environment.autoescape
        self._default_autoescape = default_autoescape

    @internalcode
    @evalcontextfunction
    def __call__(self, *args, **kwargs):
        # This requires a bit of explanation,  In the past we used to
        # decide largely based on compile-time information if a macro is
        # safe or unsafe.  While there was a volatile mode it was largely
        # unused for deciding on escaping.  This turns out to be
        # problematic for macros because whether a macro is safe depends not
        # on the escape mode when it was defined, but rather when it was used.
        #
        # Because however we export macros from the module system and
        # there are historic callers that do not pass an eval context (and
        # will continue to not pass one), we need to perform an instance
        # check here.
        #
        # This is considered safe because an eval context is not a valid
        # argument to callables otherwise anyway.  Worst case here is
        # that if no eval context is passed we fall back to the compile
        # time autoescape flag.
        if args and isinstance(args[0], EvalContext):
            autoescape = args[0].autoescape
            args = args[1:]
        else:
            autoescape = self._default_autoescape

        # try to consume the positional arguments
        arguments = list(args[: self._argument_count])
        off = len(arguments)

        # For information why this is necessary refer to the handling
        # of caller in the `macro_body` handler in the compiler.
        found_caller = False

        # if the number of arguments consumed is not the number of
        # arguments expected we start filling in keyword arguments
        # and defaults.
        if off != self._argument_count:
            for name in self.arguments[len(arguments) :]:
                try:
                    value = kwargs.pop(name)
                except KeyError:
                    value = missing
                if name == "caller":
                    found_caller = True
                arguments.append(value)
        else:
            found_caller = self.explicit_caller

        # it's important that the order of these arguments does not change
        # if not also changed in the compiler's `function_scoping` method.
        # the order is caller, keyword arguments, positional arguments!
        if self.caller and not found_caller:
            caller = kwargs.pop("caller", None)
            if caller is None:
                caller = self._environment.undefined("No caller defined", name="caller")
            arguments.append(caller)

        if self.catch_kwargs:
            arguments.append(kwargs)
        elif kwargs:
            if "caller" in kwargs:
                raise TypeError(
                    "macro %r was invoked with two values for "
                    "the special caller argument.  This is "
                    "most likely a bug." % self.name
                )
            raise TypeError(
                "macro %r takes no keyword argument %r"
                % (self.name, next(iter(kwargs)))
            )
        if self.catch_varargs:
            arguments.append(args[self._argument_count :])
        elif len(args) > self._argument_count:
            raise TypeError(
                "macro %r takes not more than %d argument(s)"
                % (self.name, len(self.arguments))
            )

        return self._invoke(arguments, autoescape)

    def _invoke(self, arguments, autoescape):
        """This method is being swapped out by the async implementation."""
        rv = self._func(*arguments)
        if autoescape:
            rv = Markup(rv)
        return rv

    def __repr__(self):
        return "<%s %s>" % (
            self.__class__.__name__,
            self.name is None and "anonymous" or repr(self.name),
        )


@implements_to_string
class Undefined(object):
    """The default undefined type.  This undefined type can be printed and
    iterated over, but every other access will raise an :exc:`UndefinedError`:

    >>> foo = Undefined(name='foo')
    >>> str(foo)
    ''
    >>> not foo
    True
    >>> foo + 42
    Traceback (most recent call last):
      ...
    jinja2.exceptions.UndefinedError: 'foo' is undefined
    """

    __slots__ = (
        "_undefined_hint",
        "_undefined_obj",
        "_undefined_name",
        "_undefined_exception",
    )

    def __init__(self, hint=None, obj=missing, name=None, exc=UndefinedError):
        self._undefined_hint = hint
        self._undefined_obj = obj
        self._undefined_name = name
        self._undefined_exception = exc

    @property
    def _undefined_message(self):
        """Build a message about the undefined value based on how it was
        accessed.
        """
        if self._undefined_hint:
            return self._undefined_hint

        if self._undefined_obj is missing:
            return "%r is undefined" % self._undefined_name

        if not isinstance(self._undefined_name, string_types):
            return "%s has no element %r" % (
                object_type_repr(self._undefined_obj),
                self._undefined_name,
            )

        return "%r has no attribute %r" % (
            object_type_repr(self._undefined_obj),
            self._undefined_name,
        )

    @internalcode
    def _fail_with_undefined_error(self, *args, **kwargs):
        """Raise an :exc:`UndefinedError` when operations are performed
        on the undefined value.
        """
        raise self._undefined_exception(self._undefined_message)

    @internalcode
    def __getattr__(self, name):
        if name[:2] == "__":
            raise AttributeError(name)
        return self._fail_with_undefined_error()

    __add__ = (
        __radd__
    ) = (
        __mul__
    ) = (
        __rmul__
    ) = (
        __div__
    ) = (
        __rdiv__
    ) = (
        __truediv__
    ) = (
        __rtruediv__
    ) = (
        __floordiv__
    ) = (
        __rfloordiv__
    ) = (
        __mod__
    ) = (
        __rmod__
    ) = (
        __pos__
    ) = (
        __neg__
    ) = (
        __call__
    ) = (
        __getitem__
    ) = (
        __lt__
    ) = (
        __le__
    ) = (
        __gt__
    ) = (
        __ge__
    ) = (
        __int__
    ) = (
        __float__
    ) = (
        __complex__
    ) = __pow__ = __rpow__ = __sub__ = __rsub__ = _fail_with_undefined_error

    def __eq__(self, other):
        return type(self) is type(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(type(self))

    def __str__(self):
        return u""

    def __len__(self):
        return 0

    def __iter__(self):
        if 0:
            yield None

    def __nonzero__(self):
        return False

    __bool__ = __nonzero__

    def __repr__(self):
        return "Undefined"


def make_logging_undefined(logger=None, base=None):
    """Given a logger object this returns a new undefined class that will
    log certain failures.  It will log iterations and printing.  If no
    logger is given a default logger is created.

    Example::

        logger = logging.getLogger(__name__)
        LoggingUndefined = make_logging_undefined(
            logger=logger,
            base=Undefined
        )

    .. versionadded:: 2.8

    :param logger: the logger to use.  If not provided, a default logger
                   is created.
    :param base: the base class to add logging functionality to.  This
                 defaults to :class:`Undefined`.
    """
    if logger is None:
        import logging

        logger = logging.getLogger(__name__)
        logger.addHandler(logging.StreamHandler(sys.stderr))
    if base is None:
        base = Undefined

    def _log_message(undef):
        if undef._undefined_hint is None:
            if undef._undefined_obj is missing:
                hint = "%s is undefined" % undef._undefined_name
            elif not isinstance(undef._undefined_name, string_types):
                hint = "%s has no element %s" % (
                    object_type_repr(undef._undefined_obj),
                    undef._undefined_name,
                )
            else:
                hint = "%s has no attribute %s" % (
                    object_type_repr(undef._undefined_obj),
                    undef._undefined_name,
                )
        else:
            hint = undef._undefined_hint
        logger.warning("Template variable warning: %s", hint)

    class LoggingUndefined(base):
        def _fail_with_undefined_error(self, *args, **kwargs):
            try:
                return base._fail_with_undefined_error(self, *args, **kwargs)
            except self._undefined_exception as e:
                logger.error("Template variable error: %s", str(e))
                raise e

        def __str__(self):
            rv = base.__str__(self)
            _log_message(self)
            return rv

        def __iter__(self):
            rv = base.__iter__(self)
            _log_message(self)
            return rv

        if PY2:

            def __nonzero__(self):
                rv = base.__nonzero__(self)
                _log_message(self)
                return rv

            def __unicode__(self):
                rv = base.__unicode__(self)
                _log_message(self)
                return rv

        else:

            def __bool__(self):
                rv = base.__bool__(self)
                _log_message(self)
                return rv

    return LoggingUndefined


# No @implements_to_string decorator here because __str__
# is not overwritten from Undefined in this class.
# This would cause a recursion error in Python 2.
class ChainableUndefined(Undefined):
    """An undefined that is chainable, where both ``__getattr__`` and
    ``__getitem__`` return itself rather than raising an
    :exc:`UndefinedError`.

    >>> foo = ChainableUndefined(name='foo')
    >>> str(foo.bar['baz'])
    ''
    >>> foo.bar['baz'] + 42
    Traceback (most recent call last):
      ...
    jinja2.exceptions.UndefinedError: 'foo' is undefined

    .. versionadded:: 2.11.0
    """

    __slots__ = ()

    def __html__(self):
        return self.__str__()

    def __getattr__(self, _):
        return self

    __getitem__ = __getattr__


@implements_to_string
class DebugUndefined(Undefined):
    """An undefined that returns the debug info when printed.

    >>> foo = DebugUndefined(name='foo')
    >>> str(foo)
    '{{ foo }}'
    >>> not foo
    True
    >>> foo + 42
    Traceback (most recent call last):
      ...
    jinja2.exceptions.UndefinedError: 'foo' is undefined
    """

    __slots__ = ()

    def __str__(self):
        if self._undefined_hint is None:
            if self._undefined_obj is missing:
                return u"{{ %s }}" % self._undefined_name
            return "{{ no such element: %s[%r] }}" % (
                object_type_repr(self._undefined_obj),
                self._undefined_name,
            )
        return u"{{ undefined value printed: %s }}" % self._undefined_hint


@implements_to_string
class StrictUndefined(Undefined):
    """An undefined that barks on print and iteration as well as boolean
    tests and all kinds of comparisons.  In other words: you can do nothing
    with it except checking if it's defined using the `defined` test.

    >>> foo = StrictUndefined(name='foo')
    >>> str(foo)
    Traceback (most recent call last):
      ...
    jinja2.exceptions.UndefinedError: 'foo' is undefined
    >>> not foo
    Traceback (most recent call last):
      ...
    jinja2.exceptions.UndefinedError: 'foo' is undefined
    >>> foo + 42
    Traceback (most recent call last):
      ...
    jinja2.exceptions.UndefinedError: 'foo' is undefined
    """

    __slots__ = ()
    __iter__ = (
        __str__
    ) = (
        __len__
    ) = (
        __nonzero__
    ) = __eq__ = __ne__ = __bool__ = __hash__ = Undefined._fail_with_undefined_error


# remove remaining slots attributes, after the metaclass did the magic they
# are unneeded and irritating as they contain wrong data for the subclasses.
del (
    Undefined.__slots__,
    ChainableUndefined.__slots__,
    DebugUndefined.__slots__,
    StrictUndefined.__slots__,
)



"""
test package backported for python-future.

Its primary purpose is to allow use of "import test.support" for running
the Python standard library unit tests using the new Python 3 stdlib
import location.

Python 3 renamed test.test_support to test.support.
"""



import json

from old_helper import BaseRepository
from typing import List, Optional


class OrderRepository(BaseRepository):
    def __init__(self, mb_context):
        super(OrderRepository, self).__init__(mb_context)

    def get_orders_benefit(self, initial_date, end_date, operator_id, pos_list):
        # type: (str, str, Optional[str], List[int]) -> List

        def inner_func(conn):

            query_operator = " and SessionId like '%user={0},%'".format(operator_id) if operator_id is not None else ""
            query = """SELECT o.OrderId, ocp.Value as BenefitList
                         FROM Orders o
                        INNER JOIN OrderTender ot
                           ON o.OrderId = ot.OrderId
                        INNER JOIN OrderCustomProperties ocp
                           ON ocp.OrderId = o.OrderId AND ocp.Key = 'BENEFIT_LIST'
                        WHERE o.BusinessPeriod >= '{0}' AND o.BusinessPeriod <= '{1}' AND o.StateId = 5
                        {2}
                        GROUP by o.OrderId
                        """.format(initial_date, end_date, query_operator)

            benefits = []
            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    benefit_list_json = json.loads(row.get_entry("BenefitList"))
                    benefit_list_json = json.loads(benefit_list_json.get("benefitList", None))
                    for benefit in benefit_list_json:
                        benefit_applied = benefit.get("benefit_applied", False)
                        benefit_total_value = benefit.get("benefit_total_value", 0)
                        if not benefit_applied or benefit_total_value == 0:
                            continue
                        benefits.append(benefit)

            return benefits

        return self.execute_in_all_databases_returning_flat_list(inner_func, pos_list)
